# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Lights.

Light objects in a `Scene`.

See Also:
    Tutorials:

    - :doc:`examples/00-Basic-tutorials/04-Scene-properties`
    - :doc:`examples/00-Basic-tutorials/05-Lighting-setups`
"""

import math

from . import _common


class Light(object):
    """A light.

    Args:
        direction ((3, ) `numpy.ndarray` of ``float32``):
            Vector direction the light points (in `Camera` space).

        color ((3, ) `numpy.ndarray` of ``float32``):
            Linear RGB color and intensity of the light.

        theta (float): Half angle of the cone that defines the area of the light
            (in radians).

    In fresnel, lights are area lights at an infinite distance away in the
    given *direction* and are circular with the size set by *theta*. *color*
    sets the light intensity. A ``(0.5, 0.5, 0.5)`` light is twice as bright as
    ``(0.25, 0.25, 0.25)``. Lights are normalized so that ``color = (1, 1, 1)``
    should provide approximately a correct exposure. Color values greater than
    1 are allowed.

    Note:
        *direction* is in `Camera` space. A direction of (1, 0, 0) sets a light
        coming from the right in the image, regardless of the camera
        position.
    """

    def __init__(self, direction, color=(1, 1, 1), theta=0.375):
        self.direction = tuple(direction)
        self.color = tuple(color)
        self.theta = float(theta)

    def __str__(self):
        """Human readable representation of Light."""
        return "<Light object with direction {0} and color {1}>".format(
            self.direction, self.color)


class _LightProxy(object):

    def __init__(self, light_list, idx):
        self._light_list = light_list
        self._idx = idx

    @property
    def direction(self):
        v = self._light_list._lights.getDirection(self._idx)
        d = (v.x, v.y, v.z)
        return d

    @direction.setter
    def direction(self, v):
        self._light_list._lights.setDirection(self._idx, _common.vec3f(*v))

    @property
    def color(self):
        v = self._light_list._lights.getColor(self._idx)
        c = (v.r, v.g, v.b)
        return c

    @color.setter
    def color(self, v):
        self._light_list._lights.setColor(self._idx, _common.RGBf(*v))

    @property
    def theta(self):
        return self._light_list._lights.getTheta(self._idx)

    @theta.setter
    def theta(self, v):
        self._light_list._lights.setTheta(self._idx, v)

    def __str__(self):
        return "<Light proxy object with direction {0} and color {1}>".format(
            self.direction, self.color)


class _LightListProxy(object):

    def __init__(self, _lights=None):
        if _lights is None:
            self._lights = _common.Lights()
            self._lights.N = 0
        else:
            self._lights = _lights

    def __len__(self):
        return self._lights.N

    def clear(self):
        self._lights.N = 0

    def append(self, light):
        if len(self) >= 4:
            raise IndexError("Cannot add more than 4 lights")

        i = self._lights.N
        self._lights.setDirection(i, _common.vec3f(*light.direction))
        self._lights.setColor(i, _common.RGBf(*light.color))
        self._lights.setTheta(i, light.theta)
        self._lights.N = i + 1

    def __getitem__(self, idx):
        if idx >= self._lights.N:
            raise IndexError("Indexing past the end of the list")

        return _LightProxy(self, idx)


def butterfly():
    """Create a butterfly lighting setup.

    The butterfly portrait lighting setup is front lighting with the key light
    (index 0) placed high above the camera and the fill light (index 1) below
    the camera.

    Returns:
        list[Light]: The lights.
    """
    res = []
    theta1 = 50 * math.pi / 180
    res.append(
        Light(direction=(0, math.sin(theta1), math.cos(theta1)),
              color=(1.0, 1.0, 1.0),
              theta=math.pi / 4))
    theta2 = -30 * math.pi / 180
    res.append(
        Light(direction=(0, math.sin(theta2), math.cos(theta2)),
              color=(0.1, 0.1, 0.1),
              theta=math.pi / 2))
    return res


def loop(side='right'):
    """Create a loop lighting setup.

    The loop portrait lighting setup places the key light slightly to one side
    of the camera and slightly up (index 0). The fill light is on the other side
    of the camera at the level of the camera (index 1).

    Args:
        side (str): 'right' or 'left' to choose which side of the camera to
            place the key light.

    Returns:
        list[Light]: The lights.
    """
    sign = {'right': 1, 'left': -1}

    res = []
    phi1 = sign[side] * 25 * math.pi / 180
    theta1 = (90 - 20) * math.pi / 180
    res.append(
        Light(direction=(math.sin(theta1) * math.sin(phi1), math.cos(theta1),
                         math.sin(theta1) * math.cos(phi1)),
              color=(1.0, 1.0, 1.0),
              theta=math.pi / 4))
    phi1 = -sign[side] * 40 * math.pi / 180
    theta1 = (90) * math.pi / 180
    res.append(
        Light(direction=(math.sin(theta1) * math.sin(phi1), math.cos(theta1),
                         math.sin(theta1) * math.cos(phi1)),
              color=(0.1, 0.1, 0.1),
              theta=math.pi / 2))
    return res


def rembrandt(side='right'):
    """Create a Rembrandt lighting setup.

    The Rembrandt portrait lighting setup places the key light  45 degrees to
    one side of the camera and slightly up (index 0). The fill light is on the
    other side of the camera at the level of the camera (index 1).

    Args:
        side (str): 'right' or 'left' to choose which side of the camera to
            place the key light.

    Returns:
        list[Light]: The lights.
    """
    sign = {'right': 1, 'left': -1}

    res = []
    phi1 = sign[side] * 45 * math.pi / 180
    theta1 = (90 - 20) * math.pi / 180
    res.append(
        Light(direction=(math.sin(theta1) * math.sin(phi1), math.cos(theta1),
                         math.sin(theta1) * math.cos(phi1)),
              color=(1.0, 1.0, 1.0),
              theta=math.pi / 4))
    phi1 = -sign[side] * 45 * math.pi / 180
    theta1 = (90) * math.pi / 180
    res.append(
        Light(direction=(math.sin(theta1) * math.sin(phi1), math.cos(theta1),
                         math.sin(theta1) * math.cos(phi1)),
              color=(0.1, 0.1, 0.1),
              theta=math.pi / 2))
    return res


def lightbox():
    """Create a light box lighting setup.

    The light box lighting setup places a single area light that covers the top,
    bottom, left, and right. Use `Path` tracing for best results with this
    setup.

    Returns:
        list[Light]: The lights.
    """
    res = []
    res.append(Light(direction=(0, 0, 1), color=(0.9, 0.9, 0.9), theta=math.pi))
    return res


def cloudy():
    """Create a cloudy day lighting setup.

    The cloudy lighting setup mimics a cloudy day. A strong light comes from all
    directions above. A weaker light comes from all directions below (accounting
    for light reflected off the ground). Use `Path` tracing for best results
    with this setup.

    Returns:
        list[Light]: The lights.
    """
    res = []
    res.append(
        Light(direction=(0, 1, 0), color=(0.9, 0.9, 0.9), theta=math.pi / 2))
    res.append(
        Light(direction=(0, -1, 0), color=(0.1, 0.1, 0.1), theta=math.pi / 2))
    return res


def ring():
    """Create a ring lighting setup.

    The ring lighting setup provides a strong front area light. This type of
    lighting is common in fashion photography. Use `Path` tracing for best
    results with this setup.

    Returns:
        list[Light]: The lights.
    """
    res = []
    res.append(
        Light(direction=(0, 0, 1), color=(0.9, 0.9, 0.9), theta=math.pi / 4))
    return res



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Materials describe the way light interacts with surfaces."""

from . import _common


class Material(object):
    """Define material properties.

    Materials control how light interacts with the geometry.

    Args:
        solid (float): Set to 1 to pass through a solid color, regardless of the
            light and view angle.

        color ((3, ) `numpy.ndarray` of ``float32``)): Linear material color.

        primitive_color_mix (float): Set to 1 to use the color provided in the
            `Geometry`, 0 to use the color specified in the `Material`, or a
            value in the range [0, 1] to mix the two colors.

        roughness (float): Roughness of the material. Nominally in the range
            [0.1, 1].

        specular (float): Control the strength of the specular highlights.
            Nominally in the range [0, 1].

        spec_trans (float): Control the amount of specular light transmission.
            In the range [0, 1].

        metal (float): Set to 0 for dielectric material, or 1 for metal.
            Intermediate values interpolate between the two.

    See Also:
        Tutorials:

        - :doc:`examples/00-Basic-tutorials/02-Material-properties`

    Note:
        Colors are in the linearized color space. Use `fresnel.color.linear` to
        convert standard sRGB colors into this space.
    """

    def __init__(self,
                 solid=0,
                 color=(0.9, 0.9, 0.9),
                 primitive_color_mix=0,
                 roughness=0.3,
                 specular=0.5,
                 spec_trans=0,
                 metal=0):
        self._material = _common.Material()

        self.solid = solid
        self.color = color
        self.roughness = roughness
        self.specular = specular
        self.metal = metal
        self.spec_trans = spec_trans
        self.primitive_color_mix = primitive_color_mix

    @property
    def solid(self):
        """float: Is this material a solid color?

        Set to 1 to pass through a solid color, regardless of the light and view
        angle.
        """
        return self._material.solid

    @solid.setter
    def solid(self, value):
        self._material.solid = float(value)

    @property
    def primitive_color_mix(self):
        """float: Mix the material color with the geometry.

        Set to 1 to use the color provided in the `Geometry`, 0 to use the color
        specified in the `Material`, or a value in the range [0, 1] to mix
        the two colors.
        """
        return self._material.primitive_color_mix

    @primitive_color_mix.setter
    def primitive_color_mix(self, value):
        self._material.primitive_color_mix = float(value)

    @property
    def color(self):
        """((3, ) `numpy.ndarray` of ``float32``)): - Linear material color."""
        return (self._material.color.r, self._material.color.g,
                self._material.color.b)

    @color.setter
    def color(self, value):
        if len(value) != 3:
            raise ValueError("colors must have length 3")
        self._material.color = _common.RGBf(*value)

    @property
    def roughness(self):
        """float: Roughness of the material.

        Nominally in the range [0.1, 1].
        """
        return self._material.roughness

    @roughness.setter
    def roughness(self, value):
        self._material.roughness = float(value)

    @property
    def specular(self):
        """float: Control the strength of the specular highlights.

        Nominally in the range [0, 1].
        """
        return self._material.specular

    @specular.setter
    def specular(self, value):
        self._material.specular = float(value)

    @property
    def spec_trans(self):
        """float: Control the amount of specular light transmission.

        In the range [0, 1].
        """
        return self._material.spec_trans

    @spec_trans.setter
    def spec_trans(self, value):
        self._material.spec_trans = float(value)

    @property
    def metal(self):
        """float: Set to 0 for dielectric material, or 1 for metal.

        Intermediate values interpolate between the two.
        """
        return self._material.metal

    @metal.setter
    def metal(self, value):
        self._material.metal = float(value)

    def _get_cpp_material(self):
        return self._material


class _MaterialProxy(object):
    """Proxy `Material` attached to a `Geometry`."""

    def __init__(self, geometry):
        self._geometry = geometry._geometry

    @property
    def solid(self):
        m = self._geometry.getMaterial()
        return m.solid

    @solid.setter
    def solid(self, value):
        m = self._geometry.getMaterial()
        m.solid = float(value)
        self._geometry.setMaterial(m)

    @property
    def primitive_color_mix(self):
        m = self._geometry.getMaterial()
        return m.primitive_color_mix

    @primitive_color_mix.setter
    def primitive_color_mix(self, value):
        m = self._geometry.getMaterial()
        m.primitive_color_mix = float(value)
        self._geometry.setMaterial(m)

    @property
    def color(self):
        m = self._geometry.getMaterial()
        return (m.color.r, m.color.g, m.color.b)

    @color.setter
    def color(self, value):
        if len(value) != 3:
            raise ValueError("colors must have length 3")

        m = self._geometry.getMaterial()
        m.color = _common.RGBf(*value)
        self._geometry.setMaterial(m)

    @property
    def roughness(self):
        m = self._geometry.getMaterial()
        return m.roughness

    @roughness.setter
    def roughness(self, value):
        m = self._geometry.getMaterial()
        m.roughness = float(value)
        self._geometry.setMaterial(m)

    @property
    def specular(self):
        m = self._geometry.getMaterial()
        return m.specular

    @specular.setter
    def specular(self, value):
        m = self._geometry.getMaterial()
        m.specular = float(value)
        self._geometry.setMaterial(m)

    @property
    def spec_trans(self):
        m = self._geometry.getMaterial()
        return m.spec_trans

    @spec_trans.setter
    def spec_trans(self, value):
        m = self._geometry.getMaterial()
        m.spec_trans = float(value)
        self._geometry.setMaterial(m)

    @property
    def metal(self):
        m = self._geometry.getMaterial()
        return m.metal

    @metal.setter
    def metal(self, value):
        m = self._geometry.getMaterial()
        m.metal = float(value)
        self._geometry.setMaterial(m)

    def _get_cpp_material(self):
        return self._geometry.getMaterial()


class _OutlineMaterialProxy(object):
    """Proxy outline `Material` attached to a Geometry."""

    def __init__(self, geometry):
        self._geometry = geometry._geometry

    @property
    def solid(self):
        m = self._geometry.getOutlineMaterial()
        return m.solid

    @solid.setter
    def solid(self, value):
        m = self._geometry.getOutlineMaterial()
        m.solid = float(value)
        self._geometry.setOutlineMaterial(m)

    @property
    def primitive_color_mix(self):
        m = self._geometry.getOutlineMaterial()
        return m.primitive_color_mix

    @primitive_color_mix.setter
    def primitive_color_mix(self, value):
        m = self._geometry.getOutlineMaterial()
        m.primitive_color_mix = float(value)
        self._geometry.setOutlineMaterial(m)

    @property
    def color(self):
        m = self._geometry.getOutlineMaterial()
        return (m.color.r, m.color.g, m.color.b)

    @color.setter
    def color(self, value):
        if len(value) != 3:
            raise ValueError("colors must have length 3")

        m = self._geometry.getOutlineMaterial()
        m.color = _common.RGBf(*value)
        self._geometry.setOutlineMaterial(m)

    @property
    def roughness(self):
        m = self._geometry.getOutlineMaterial()
        return m.roughness

    @roughness.setter
    def roughness(self, value):
        m = self._geometry.getOutlineMaterial()
        m.roughness = float(value)
        self._geometry.setOutlineMaterial(m)

    @property
    def specular(self):
        m = self._geometry.getOutlineMaterial()
        return m.specular

    @specular.setter
    def specular(self, value):
        m = self._geometry.getOutlineMaterial()
        m.specular = float(value)
        self._geometry.setOutlineMaterial(m)

    @property
    def spec_trans(self):
        m = self._geometry.getOutlineMaterial()
        return m.spec_trans

    @spec_trans.setter
    def spec_trans(self, value):
        m = self._geometry.getOutlineMaterial()
        m.spec_trans = float(value)
        self._geometry.setOutlineMaterial(m)

    @property
    def metal(self):
        m = self._geometry.getOutlineMaterial()
        return m.metal

    @metal.setter
    def metal(self, value):
        m = self._geometry.getOutlineMaterial()
        m.metal = float(value)
        self._geometry.setOutlineMaterial(m)

    def _get_cpp_material(self):
        return self._geometry.getOutlineMaterial()



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test lights."""

import fresnel
import pytest


def test_lightlist(device_):
    """Test that lightlist allows getting and setting lights."""
    scene = fresnel.Scene(device_, lights=[])
    l = scene.lights  # noqa: E741 - allow l as a name
    assert (len(l) == 0)
    l.append(
        fresnel.light.Light(direction=(1, 2, 3),
                            color=(0.5, 0.125, 0.75),
                            theta=2))
    assert (len(l) == 1)
    l.append(
        fresnel.light.Light(direction=(5, 10, 2), color=(10, 18, 6), theta=1))
    assert (len(l) == 2)
    l.append(
        fresnel.light.Light(direction=(4, 8, 16),
                            color=(100, 1200, 7),
                            theta=0.5))
    assert (len(l) == 3)
    l.append(
        fresnel.light.Light(direction=(5, 1, 2), color=(16, 32, 64), theta=.75))
    assert (len(l) == 4)

    with pytest.raises(Exception):
        l.append(fresnel.light.Light(direction=(5, 1, 2), color=(16, 32, 64)))

    assert (l[0].direction == (1, 2, 3))
    assert (l[0].color == (0.5, 0.125, 0.75))
    assert (l[0].theta == 2)

    assert (l[1].direction == (5, 10, 2))
    assert (l[1].color == (10, 18, 6))
    assert (l[1].theta == 1)

    assert (l[2].direction == (4, 8, 16))
    assert (l[2].color == (100, 1200, 7))
    assert (l[2].theta == 0.5)

    assert (l[3].direction == (5, 1, 2))
    assert (l[3].color == (16, 32, 64))
    assert (l[3].theta == 0.75)

    with pytest.raises(Exception):
        l[4].direction

    l[0].direction = (10, 1, 18)
    l[0].color = (5, 10, 20)
    l[0].theta = 0.75
    assert (l[0].direction == (10, 1, 18))
    assert (l[0].color == (5, 10, 20))
    assert (l[0].theta == 0.75)

    for light in l:
        d = light.direction
        light.direction = (-d[0], -d[1], -d[2])

    assert (l[0].direction == (-10, -1, -18))
    assert (l[1].direction == (-5, -10, -2))
    assert (l[2].direction == (-4, -8, -16))
    assert (l[3].direction == (-5, -1, -2))

    l.clear()
    assert (len(l) == 0)

    scene.lights = [
        fresnel.light.Light(direction=(1, 2, 3)),
        fresnel.light.Light(direction=(10, 20, 30),
                            color=(1, 0, 0),
                            theta=1.125)
    ]
    assert (l[0].direction == (1, 2, 3))
    assert (l[0].color == (1, 1, 1))
    assert (l[0].theta == 0.375)
    assert (l[1].direction == (10, 20, 30))
    assert (l[1].color == (1, 0, 0))
    assert (l[1].theta == 1.125)

    with pytest.raises(Exception):
        scene.lights = [
            fresnel.light.Light(direction=(1, 2, 3)),
            fresnel.light.Light(direction=(10, 20, 30)),
            fresnel.light.Light(direction=(1, 2, 3)),
            fresnel.light.Light(direction=(1, 2, 3)),
            fresnel.light.Light(direction=(1, 2, 3))
        ]



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test the Sphere geometry."""

import fresnel
import numpy
from collections import namedtuple
import PIL
import conftest
import pytest
import os
import pathlib

dir_path = pathlib.Path(os.path.realpath(__file__)).parent


def scene_four_spheres(device):
    """Create a test scene with four spheres."""
    scene = fresnel.Scene(device, lights=conftest.test_lights())

    mat = fresnel.material.Material(
        color=fresnel.color.linear([0.42, 0.267, 1]))
    fresnel.geometry.Sphere(scene,
                            position=[[1, 0, 1], [1, 0, -1], [-1, 0, 1],
                                      [-1, 0, -1]],
                            radius=1.0,
                            material=mat,
                            color=[[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1]])

    scene.camera = fresnel.camera.Orthographic(position=(10, 10, 10),
                                               look_at=(0, 0, 0),
                                               up=(0, 1, 0),
                                               height=4)

    return scene


@pytest.fixture(scope='function')
def scene_four_spheres_(device_):
    """Pytest fixture to create a test scene."""
    return scene_four_spheres(device_)


def test_render(scene_four_spheres_, generate=False):
    """Test that spheres render properly."""
    buf_proxy = fresnel.preview(scene_four_spheres_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_sphere.test_render.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_sphere.test_render.png')


def test_radius(scene_four_spheres_, generate=False):
    """Test the radius property."""
    geometry = scene_four_spheres_.geometry[0]

    r = numpy.array([0.5, 0.6, 0.8, 1.0], dtype=numpy.float32)
    geometry.radius[:] = r
    numpy.testing.assert_array_equal(r, geometry.radius[:])

    buf_proxy = fresnel.preview(scene_four_spheres_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_sphere.test_radius.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_sphere.test_radius.png')


def test_position(scene_four_spheres_, generate=False):
    """Test the position property."""
    geometry = scene_four_spheres_.geometry[0]

    p = numpy.array([[1.5, 0, 1], [1.5, 0, -1], [-1.5, 0, 1], [-1.5, 0, -1]],
                    dtype=numpy.float32)
    geometry.position[:] = p
    numpy.testing.assert_array_equal(p, geometry.position[:])

    buf_proxy = fresnel.preview(scene_four_spheres_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_sphere.test_position.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_sphere.test_position.png')


def test_color(scene_four_spheres_, generate=False):
    """Test the color property."""
    geometry = scene_four_spheres_.geometry[0]
    geometry.material.primitive_color_mix = 1.0

    c = fresnel.color.linear(
        numpy.array([[1, 1, 1], [0, 0, 1], [0, 1, 0], [1, 0, 0]],
                    dtype=numpy.float32))
    geometry.color[:] = c
    numpy.testing.assert_array_equal(c, geometry.color[:])

    buf_proxy = fresnel.preview(scene_four_spheres_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_sphere.test_color.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_sphere.test_color.png')


def test_outline(scene_four_spheres_, generate=False):
    """Test that outlines render properly."""
    geometry = scene_four_spheres_.geometry[0]
    geometry.outline_width = 0.1

    buf_proxy = fresnel.preview(scene_four_spheres_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_sphere.test_outline.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_sphere.test_outline.png')


if __name__ == '__main__':
    struct = namedtuple("struct", "param")
    device = conftest.device(struct(('cpu', None)))

    scene = scene_four_spheres(device)
    test_render(scene, generate=True)

    scene = scene_four_spheres(device)
    test_radius(scene, generate=True)

    scene = scene_four_spheres(device)
    test_position(scene, generate=True)

    scene = scene_four_spheres(device)
    test_color(scene, generate=True)

    scene = scene_four_spheres(device)
    test_outline(scene, generate=True)



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test the Box geometry."""

import os
import pathlib
from collections import namedtuple

import numpy
import PIL
import pytest

import conftest
import fresnel

dir_path = pathlib.Path(os.path.realpath(__file__)).parent


def scene_box(device):
    """Create a test scene."""
    scene = fresnel.Scene(device, lights=conftest.test_lights())

    fresnel.geometry.Box(scene, [1, 2, 3, 0.4, 0.5, 0.6],
                         box_radius=0.2,
                         box_color=[1, 0, 1])

    scene.camera = fresnel.camera.Orthographic(position=(10, 10, 10),
                                               look_at=(0, 0, 0),
                                               up=(0, 1, 0),
                                               height=4)

    return scene


@pytest.fixture(scope="function")
def scene_box_(device_):
    """Pytest fixture to create a test scene."""
    return scene_box(device_)


def test_render(scene_box_, generate=False):
    """Test that Box renders properly."""
    buf_proxy = fresnel.preview(scene_box_, w=150, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode="RGBA").save(
            open("output/test_geometry_box.test_render.png", "wb"), "png")
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / "reference" / "test_geometry_box.test_render.png")


def test_radius(scene_box_, generate=False):
    """Test the radius property."""
    geometry = scene_box_.geometry[0]

    r = numpy.array(0.1, dtype=numpy.float32)
    geometry.radius[:] = r
    numpy.testing.assert_array_equal(r, geometry.radius[:])

    buf_proxy = fresnel.preview(scene_box_, w=150, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode="RGBA").save(
            open("output/test_geometry_box.test_radius.png", "wb"), "png")
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / "reference" / "test_geometry_box.test_radius.png")


def test_color(scene_box_, generate=False):
    """Test the color property."""
    geometry = scene_box_.geometry[0]
    geometry.material.primitive_color_mix = 1.0

    c = numpy.zeros((12, 2, 3), dtype=numpy.float32) + [1, 0, 0]
    geometry.color[:] = c
    numpy.testing.assert_array_equal(c, geometry.color[:])

    buf_proxy = fresnel.preview(scene_box_, w=150, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode="RGBA").save(
            open("output/test_geometry_box.test_color.png", "wb"), "png")
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / "reference" / "test_geometry_box.test_color.png")


def test_box_radius(scene_box_, generate=False):
    """Test the box_radius property."""
    geometry = scene_box_.geometry[0]

    r = 0.1
    geometry.box_radius = r
    assert r == pytest.approx(geometry.box_radius)

    buf_proxy = fresnel.preview(scene_box_, w=150, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode="RGBA").save(
            open("output/test_geometry_box.test_box_radius.png", "wb"), "png")
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / "reference" / "test_geometry_box.test_box_radius.png")


def test_box_color(scene_box_, generate=False):
    """Test the box_color property."""
    geometry = scene_box_.geometry[0]

    c = numpy.array([1, 0, 1], dtype=numpy.float32)
    geometry.box_color = c
    numpy.testing.assert_array_equal(c, geometry.box_color)

    buf_proxy = fresnel.preview(scene_box_, w=150, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode="RGBA").save(
            open("output/test_geometry_box.test_box_color.png", "wb"), "png")
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / "reference" / "test_geometry_box.test_box_color.png")


def test_box_update(scene_box_, generate=False):
    """Test the box property."""
    box_tuple = namedtuple("box_tuple", "Lx Ly Lz xy xz yz")
    box_list = [
        2,
        [1.5],
        [1, 2, 3],
        [1, 1.2, 2, 0.2, 0, 0.9],
        box_tuple(2, 2, 3, 0.5, 0.5, 0.7),
        {
            "Lx": 1,
            "Ly": 2,
            "Lz": 1,
            "xy": 0.4,
            "xz": 0.5,
            "yz": 0.3
        },
    ]

    for update_box in box_list:

        geometry = scene_box_.geometry[0]

        geometry.box = update_box

        assert isinstance(geometry.box, tuple)
        assert len(geometry.box) == 6

        buf_proxy = fresnel.preview(scene_box_, w=150, h=100, anti_alias=False)

        box_index = box_list.index(update_box)

        if generate:
            PIL.Image.fromarray(buf_proxy[:], mode="RGBA").save(
                open(f"output/test_geometry_box.test_render{box_index}.png",
                     "wb"), "png")
        else:
            conftest.assert_image_approx_equal(
                buf_proxy[:], dir_path / "reference"
                / f"test_geometry_box.test_render{box_index}.png")


if __name__ == "__main__":
    struct = namedtuple("struct", "param")
    device = conftest.device(struct(("cpu", None)))

    scene = scene_box(device)
    test_render(scene, generate=True)

    scene = scene_box(device)
    test_radius(scene, generate=True)

    scene = scene_box(device)
    test_color(scene, generate=True)

    scene = scene_box(device)
    test_box_radius(scene, generate=True)

    scene = scene_box(device)
    test_box_color(scene, generate=True)

    scene = scene_box(device)
    test_box_update(scene, generate=True)



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Interactive Qt widgets."""

import sys
import numpy

# workaround bug in ipython that prevents pyside2 importing
# https://github.com/jupyter/qtconsole/pull/280
try:
    import IPython.external.qt_loaders
    if type(sys.meta_path[0]) == IPython.external.qt_loaders.ImportDenier:
        del sys.meta_path[0]
except:  # noqa
    pass

from PySide2 import QtGui
from PySide2 import QtCore
from PySide2 import QtWidgets
import rowan
import copy

from fresnel import tracer, camera

# initialize QApplication
# but not in sphinx
if 'sphinx' not in sys.modules:
    from PySide2.QtWidgets import QWidget

    app = QtCore.QCoreApplication.instance()
    if app is None:
        app = QtWidgets.QApplication(['fresnel'])
else:
    # work around bug where sphinx does not find classes that have QWidget
    # as a parent
    QWidget = object


class _CameraController3D:
    """Helper class to control camera movement.

    Args:
        camera (Camera): The camera to control.

    Note:
        Call `start` on a mouse down event and then `orbit`, `pan` or `zoom`
        on mouse move events to adjust the camera from that starting point.
    """

    def __init__(self, camera):
        self.camera = camera

    def start(self):
        """Record the initial camera position."""
        self._start_camera = copy.copy(self.camera)

    def orbit(self, yaw=0, pitch=0, roll=0, factor=-0.0025, slight=False):
        """Orbit the camera about the look_at point."""
        if slight:
            factor = factor * 0.1

        basis = numpy.array(self._start_camera.basis)

        q1 = rowan.from_axis_angle(basis[1, :], factor * yaw)
        q2 = rowan.from_axis_angle(basis[0, :], factor * pitch)
        q3 = rowan.from_axis_angle(basis[2, :], factor * roll)
        q = rowan.multiply(q2, rowan.multiply(q1, q3))

        v = self._start_camera.position - self._start_camera.look_at
        v = rowan.rotate(q, v)

        self.camera.position = self._start_camera.look_at + v
        self.camera.up = rowan.rotate(q, basis[1, :])

    def pan(self, x, y, slight=False):
        """Pan the camera parallel to the focal plane."""
        # TODO: this should be the height at the focal plane
        factor = self._start_camera.height

        if slight:
            factor = factor * 0.1

        basis = numpy.array(self._start_camera.basis)

        delta = factor * (x * basis[0, :] + y * basis[1, :])

        self.camera.position = self._start_camera.position + delta
        self.camera.look_at = self._start_camera.look_at + delta

    def zoom(self, s, slight=False):
        """Zoom the view."""
        factor = 0.0015

        if slight:
            factor = factor * 0.1

        # TODO: different types of zoom for perspective cameras
        self.camera.height = self._start_camera.height * (1 - s * factor)


class SceneView(QWidget):
    """View a fresnel Scene in real time.

    `SceneView` is a PySide2 widget that displays a `Scene`, rendering it with
    `Path` interactively. Use the mouse to rotate the camera view.

    Args:
        scene (`Scene`): The scene to display.
        max_samples (int): Sample a total of ``max_samples``.

    * Left click to pitch and yaw
    * Right click to roll and zoom
    * Middle click to pan
    * Hold ctrl to make small adjustments

    .. rubric:: Using in a standalone script

    To use SceneView in a standalone script, import the `fresnel.interact`
    module, create a `Scene`, instantiate the `SceneView`, show it, and start
    the app event loop.

    .. code-block:: python

        import fresnel, fresnel.interact
        # build scene
        view = fresnel.interact.SceneView(scene)
        view.show()
        fresnel.interact.app.exec_();

    .. rubric:: Using with Jupyter notebooks

    To use SceneView in a Jupyter notebook, import PySide2.QtCore and activate
    Jupyter's qt5 integration.

    .. code-block:: python

        from PySide2 import QtCore
        % gui qt


    Import the :py:mod:`fresnel.interact` module, create a `Scene`, and
    instantiate the `SceneView`. Do not call the app event loop, Jupyter is
    already running the event loop in the background. When the SceneView object
    is the result of a cell, it will automatically show and activate focus.

    .. code-block:: python

        import fresnel, fresnel.interact
        # build Scene
        fresnel.interact.SceneView(scene)

    Note:
        The interactive window will open on the system that *hosts* Jupyter.

    See Also:
        Tutorials:

        - :doc:`examples/02-Advanced-topics/03-Interactive-scene-view`
    """
    """Qt Signal sent when rendering starts at a new camera position."""
    rendering = QtCore.Signal(camera.Camera)

    TIMEOUT = 100
    """Timeout for delayed actions to take effect."""

    def __init__(self, scene, max_samples=2000):
        super().__init__()
        self.setWindowTitle("fresnel: scene viewer")

        self.setMinimumSize(10, 10)

        self._scene = scene
        self._max_samples = max_samples

        # fire off a timer to repaint the window as often as possible
        self._repaint_timer = QtCore.QTimer(self)
        self._repaint_timer.timeout.connect(self.update)

        # initialize a single-shot timer to delay resizing
        self._resize_timer = QtCore.QTimer(self)
        self._resize_timer.setSingleShot(True)
        self._resize_timer.timeout.connect(self._resize_done)

        # initialize the tracer
        self._tracer = tracer.Path(device=scene.device, w=10, h=10)
        self._low_res_tracer = tracer.Path(device=scene.device, w=10, h=10)
        self._is_rendering = False
        self._initial_resize = True

        # track render times
        self._frames_painted = 0

        # flag to notify view rotation
        self._camera_update_mode = None
        self._mouse_initial_pos = None
        self._render_high_res = True

        # timer to return to high res
        self._low_res_timer = QtCore.QTimer(self)
        self._low_res_timer.setSingleShot(True)
        self._low_res_timer.timeout.connect(self._low_res_done)

        self._camera_controller = _CameraController3D(self._scene.camera)
        self.ipython_display_formatter = 'text'

    @property
    def scene(self):
        """Scene: The scene rendered in this view."""
        return self._scene

    @scene.setter
    def scene(self, scene):
        self._scene = scene
        self._start_rendering()

    def _resize_done(self):
        """Resize the tracer after a delay."""
        # resize the tracer
        self._tracer.resize(w=self.width(), h=self.height())
        self._low_res_tracer.resize(w=self.width() // 4, h=self.height() // 4)
        self._start_rendering()

    def _low_res_done(self):
        """Done rendering in low resolution."""
        self._render_high_res = True

    def _stop_rendering(self):
        """Stop sampling the scene."""
        self._repaint_timer.stop()
        self._is_rendering = False

    def _start_rendering(self):
        """Start sampling the scene."""
        # send signal
        self.rendering.emit(self._scene.camera)

        self._is_rendering = True
        self._samples = 0
        self._tracer.reset()
        self._low_res_tracer.reset()
        self._repaint_timer.start()

    #####################################
    # Qt methods:

    def minimumSizeHint(self):  # noqa: N802 - allow Qt style naming
        """Specify the minimum window size hint to Qt.

        :meta private:
        """
        return QtCore.QSize(1610, 1000)

    def paintEvent(self, event):  # noqa: N802 - allow Qt style naming
        """Paint the window.

        :meta private:
        """
        if self._is_rendering:
            # Render the hi-res scene when not moving the camera
            if self._render_high_res:
                self._tracer.render(self._scene)

                self._samples += 1
                if self._samples >= self._max_samples:
                    self._stop_rendering()
            else:
                # Render the low -res scene when moving the camera
                self._low_res_tracer.render(self._scene)

        # Display the active buffer
        if self._render_high_res:
            image_array = self._tracer.output
        else:
            image_array = self._low_res_tracer.output

        # display the rendered scene in the widget
        image_array.buf.map()
        img = QtGui.QImage(image_array.buf, image_array.shape[1],
                           image_array.shape[0], QtGui.QImage.Format_RGBA8888)
        qp = QtGui.QPainter(self)
        target = QtCore.QRectF(0, 0, self.width(), self.height())
        source = QtCore.QRectF(0.0, 0.0, image_array.shape[1],
                               image_array.shape[0])

        qp.drawImage(target, img, source)
        qp.end()
        image_array.buf.unmap()

        # Display FPS
        self._frames_painted += 1

    def resizeEvent(self, event):  # noqa: N802 - allow Qt style naming
        """Adjust the size of the tracer as the window resizes.

        :meta private:
        """
        # for the initial window size, resize immediately
        if self._initial_resize:
            self._resize_done()
            self._initial_resize = False
        else:
            # otherwise, defer resizing the tracer until the window sits still
            # for a bit
            self._resize_timer.start(self.TIMEOUT)

    def mouseMoveEvent(self, event):  # noqa: N802 - allow Qt style naming
        """Respond to mouse move events.

        :meta private:
        """
        delta = event.pos() - self._mouse_initial_pos

        if self._camera_update_mode == 'pitch/yaw':
            self._camera_controller.orbit(yaw=delta.x(),
                                          pitch=delta.y(),
                                          slight=event.modifiers()
                                          & QtCore.Qt.ControlModifier)

        elif self._camera_update_mode == 'roll/zoom':
            self._camera_controller.orbit(roll=delta.x(),
                                          slight=event.modifiers()
                                          & QtCore.Qt.ControlModifier)
            self._camera_controller.zoom(-delta.y(),
                                         slight=event.modifiers()
                                         & QtCore.Qt.ControlModifier)

        elif self._camera_update_mode == 'pan':
            h = self.height()
            self._camera_controller.pan(x=-delta.x() / h,
                                        y=delta.y() / h,
                                        slight=event.modifiers()
                                        & QtCore.Qt.ControlModifier)

        self._start_rendering()
        self.update()
        event.accept()

    def mousePressEvent(self, event):  # noqa: N802 - allow Qt style naming
        """Respond to mouse press events.

        :meta private:
        """
        self._mouse_initial_pos = event.pos()
        self._camera_controller.start()
        event.accept()

        if event.button() == QtCore.Qt.LeftButton:
            self._camera_update_mode = 'pitch/yaw'
        elif event.button() == QtCore.Qt.RightButton:
            self._camera_update_mode = 'roll/zoom'
        elif event.button() == QtCore.Qt.MiddleButton:
            self._camera_update_mode = 'pan'

        self._render_high_res = False
        self._start_rendering()

    def mouseReleaseEvent(self, event):  # noqa: N802 - allow Qt style naming
        """Respond to mouse release events.

        :meta private:
        """
        if self._camera_update_mode is not None:
            self._camera_update_mode = None
            event.accept()

        self._render_high_res = True

    def wheelEvent(self, event):  # noqa: N802 - allow Qt style naming
        """Respond to mouse wheel events.

        :meta private:
        """
        self._camera_controller.start()
        self._camera_controller.zoom(event.angleDelta().y(),
                                     slight=event.modifiers()
                                     & QtCore.Qt.ControlModifier)

        self._render_high_res = False
        self._low_res_timer.start(self.TIMEOUT)
        self._start_rendering()
        event.accept()



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test the Orthographic camera."""

import fresnel
import pytest
import numpy


def test_camera_fit_no_geometry(device_):
    """Test that fit() errors when there is no geometry."""
    scene = fresnel.Scene()

    # fit cannot be called on a scene with no geometries
    with pytest.raises(ValueError):
        fresnel.camera.Orthographic.fit(scene, view='front', margin=0)


def test_camera_fit_front(device_):
    """Test that fit works with the front option."""
    scene = fresnel.Scene()

    fresnel.geometry.Sphere(scene,
                            position=[[-9, -2, 0], [-5, -1, 0], [4, 0, 0],
                                      [2, 1, 0]],
                            radius=1.0)

    cam = fresnel.camera.Orthographic.fit(scene, view='front', margin=0)
    assert cam.position[0] == -2.5
    assert cam.position[1] == -0.5
    numpy.testing.assert_array_equal(cam.look_at, (-2.5, -0.5, 0))
    assert cam.height == 5


def test_camera_fit_isometric(device_):
    """Test that fit works with the isometric option."""
    scene = fresnel.Scene()

    fresnel.geometry.Sphere(scene,
                            position=[[-9, -2, 0], [-5, -1, 0], [4, 0, 0],
                                      [2, 1, 0]],
                            radius=1.0)

    fresnel.camera.Orthographic.fit(scene, view='isometric', margin=0)
    # isometric cameras do not have a simple testable format, just test that the
    # API works


def test_scene_default_camera(device_):
    """Test that there is a default camera."""
    scene = fresnel.Scene()

    fresnel.geometry.Sphere(scene,
                            position=[[-9, -2, 0], [-5, -1, 0], [4, 0, 0],
                                      [2, 1, 0]],
                            radius=1.0)

    fresnel.preview(scene, anti_alias=False)


def test_orthographic_attributes():
    """Test all properties."""
    cam = fresnel.camera.Orthographic(position=(1, 0, 10),
                                      look_at=(1, 0, 0),
                                      up=(0, 1, 0),
                                      height=6)
    numpy.testing.assert_array_equal(cam.position, (1, 0, 10))
    numpy.testing.assert_array_equal(cam.look_at, (1, 0, 0))
    numpy.testing.assert_array_equal(cam.up, (0, 1, 0))
    assert cam.height == 6

    cam2 = eval(repr(cam))
    numpy.testing.assert_array_equal(cam2.position, (1, 0, 10))
    numpy.testing.assert_array_equal(cam2.look_at, (1, 0, 0))
    numpy.testing.assert_array_equal(cam2.up, (0, 1, 0))
    assert cam2.height == 6

    cam.position = (1, 3, 8)
    cam.look_at = (20, 5, 18)
    cam.up = (1, 2, 3)
    cam.height = 112

    numpy.testing.assert_array_equal(cam.position, (1, 3, 8))
    numpy.testing.assert_array_equal(cam.look_at, (20, 5, 18))
    numpy.testing.assert_array_equal(cam.up, (1, 2, 3))
    assert cam.height == 112



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Color utilities."""

import numpy


def linear(color):
    """Convert a sRGB color (or colors) into the linear space.

    Standard tools for working with `sRGB <https://en.wikipedia.org/wiki/SRGB>`_
    colors provide gamma corrected values. fresnel needs to perform calculations
    in a linear color space. This method converts from sRGB to the linear space.
    Use :py:func:`linear` when specifying material or particle colors with sRGB
    inputs (such as you find in a color picker).

    :py:func:`linear` accepts
    `RGBA <https://en.wikipedia.org/wiki/RGBA_color_space>`_ input (such as from
    matplotlib's `colors.to_rgba
    <https://matplotlib.org/api/_as_gen/matplotlib.colors.to_rgba.html>`_
    colormap method), but ignores the alpha channel and outputs an ``Nx3``
    array.

    Args:
        color ((3, ), (4, ), (N, 3), or (N, 4) `numpy.ndarray` of\
            ``numpy.float32``): ``RGB`` or ``RGBA`` colors.

    Color components are in the range [0,1].

    Returns:
        `numpy.ndarray` with the linearized color(s), same shape as ``color``.
    """
    c = numpy.ascontiguousarray(color)
    if c.shape == (3,) or c.shape == (4,):
        out = numpy.zeros(3, dtype=numpy.float32)
        if c[0] < 0.04045:
            out[0] = c[0] / 12.92
        else:
            out[0] = ((c[0] + 0.055) / (1.055))**2.4

        if c[1] < 0.04045:
            out[1] = c[1] / 12.92
        else:
            out[1] = ((c[1] + 0.055) / (1.055))**2.4

        if c[2] < 0.04045:
            out[2] = c[2] / 12.92
        else:
            out[2] = ((c[2] + 0.055) / (1.055))**2.4

    elif c.ndim == 2 and (c.shape[1] == 3 or c.shape[1] == 4):
        out = numpy.zeros(shape=(c.shape[0], 3), dtype=numpy.float32)
        for i in range(3):
            s = c[:, i] < 0.04045
            out[s, i] = c[s, i] / 12.92
            not_s = numpy.logical_not(s)
            out[not_s, i] = ((c[not_s, i] + 0.055) / (1.055))**2.4
    else:
        raise TypeError("color must be a length 3, 4, Nx3, or Nx4 array")

    return out



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test the Path tracer."""

import fresnel
from collections import namedtuple
import PIL
import conftest
import os
import pathlib

dir_path = pathlib.Path(os.path.realpath(__file__)).parent


def test_render(scene_hex_sphere_, generate=False):
    """Test that the path tracer renders properly."""
    tracer = fresnel.tracer.Path(device=scene_hex_sphere_.device, w=100, h=100)
    tracer.seed = 11
    buf = tracer.output[:]
    assert buf.shape == (100, 100, 4)

    buf_proxy = tracer.sample(scene_hex_sphere_, samples=64, light_samples=40)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_tracer_path.test_render.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(buf_proxy[:],
                                           dir_path / 'reference'
                                           / 'test_tracer_path.test_render.png',
                                           tolerance=16)


if __name__ == '__main__':
    struct = namedtuple("struct", "param")
    device = conftest.device(struct(('gpu', 1)))

    scene = conftest.scene_hex_sphere(device)
    test_render(scene, generate=True)



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Version and build information.

Attributes:
    version (str): fresnel package version, following semantic versioning.
"""

version = "0.13.4"



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""The fresnel ray tracing package."""

import os
import numpy

from . import geometry  # noqa: F401 - ignore unused import
from . import tracer
from . import camera
from . import color  # noqa: F401 - ignore unused import (users will use)
from . import light
from . import version  # noqa: F401 - ignore unused import (users will use)

from . import _common
if _common.cpu_built():
    from . import _cpu
if _common.gpu_built():
    from . import _gpu


class Device(object):
    """Hardware device to use for ray tracing.

    Args:
        mode (str): Specify execution mode: Valid values are ``auto``, ``gpu``,
            and ``cpu``.
        n (int): Specify the number of cpu threads / GPUs this device will use.
            *None* will use all available threads / devices.

    `Device` defines hardware device to use for ray tracing. `Scene` and
    `Tracer` instances must be attached to a `Device`. You may attach any number
    of scenes and tracers to a single `Device`.

    See Also:
        Tutorials:

        - :doc:`examples/02-Advanced-topics/01-Devices`
        - :doc:`examples/02-Advanced-topics/02-Tracer-methods`

    When mode is ``auto``, the default, `Device` will select GPU rendering if
    available and fall back on CPU rendering if not. Set mode to ``gpu`` or
    ``cpu`` to force a specific mode.

    Important:
        By default (``n==None``), this device will use all available GPUs or CPU
        cores. Set *n* to the number of GPUs or CPU cores this device should
        use. When selecting *n* GPUs, the device selects the first *n* in the
        `available_gpus` list.

    Tip:
        Use only a single `Device` to reduce memory consumption.

    The static member `available_modes` lists which modes are available. For a
    mode to be available, the corresponding module must be enabled at compile
    time. Additionally, there must be at least one GPU present for the ``gpu``
    mode to be available.

    .. code-block:: python

        >>> fresnel.Device.available_modes
        ['gpu', 'cpu', 'auto']
    """

    available_modes = []
    """list[str]: Available execution modes."""

    available_gpus = []
    """list[str]: Available GPUS."""

    def __init__(self, mode='auto', n=None):
        # determine the number of available GPUs
        num_gpus = 0
        if _common.gpu_built():
            num_gpus = _gpu.get_num_available_devices()

        # determine the selected mode
        selected_mode = ''

        if mode == 'auto':
            if num_gpus > 0:
                selected_mode = 'gpu'
            else:
                selected_mode = 'cpu'
                if not _common.cpu_built():
                    raise RuntimeError("No GPUs available AND CPU "
                                       "implementation is not compiled")

        if mode == 'gpu':
            if not _common.gpu_built():
                raise RuntimeError("GPU implementation is not compiled")
            if num_gpus == 0:
                raise RuntimeError("No GPUs are available")
            selected_mode = 'gpu'

        if mode == 'cpu':
            if not _common.cpu_built():
                raise RuntimeError("CPU implementation is not compiled")
            selected_mode = 'cpu'

        if n is None:
            thread_limit = -1
        else:
            thread_limit = int(n)

        # initialize the device
        if selected_mode == 'gpu':
            self.module = _gpu
            self._device = _gpu.Device(
                os.path.dirname(os.path.realpath(__file__)), thread_limit)
            self._mode = 'gpu'
        elif selected_mode == 'cpu':
            self.module = _cpu
            self._device = _cpu.Device(thread_limit)
            self._mode = 'cpu'
        else:
            raise ValueError("Invalid mode")

    @property
    def mode(self):
        """str: The active mode."""
        return self._mode

    def __str__(self):
        """Human readable `Device` summary."""
        return '<fresnel.Device: ' + self._device.describe() + '>'


# determine available Device modes
if _common.gpu_built():
    if _gpu.get_num_available_devices() > 0:
        Device.available_modes.append('gpu')

if _common.cpu_built():
    Device.available_modes.append('cpu')

if len(Device.available_modes) > 0:
    Device.available_modes.append('auto')

# determine available Device GPUs
if _common.gpu_built():
    gpus_str = _gpu.Device.getAllGPUs()
    gpus_list = gpus_str.split('\n')
    if len(gpus_list) >= 2:
        Device.available_gpus = gpus_list[:-1]


class Scene(object):
    """Content of the scene to ray trace.

    Args:
        device (Device): Device to use when rendering the scene.

        camera (camera.Camera): Camera to view the scene. When `None`,
          defaults to::

            camera.Orthographic(position=(0, 0, 100),
                                look_at=(0, 0, 0),
                                up=(0, 1, 0),
                                height=100)

        lights (list[Light]): Lights to light the scene. When `None`, defaults
          to: ``light.rembrandt()``

    `Scene` defines the contents of the scene to be traced, including any number
    of `Geometry` objects, the `Camera`, the `background_color`,
    `background_alpha`, and `lights`.

    Every `Scene` must be associated with a `Device`. For convenience, `Scene`
    creates a default `Device` when *device* is ``None``.

    See Also:
        Tutorials:

        - :doc:`examples/00-Basic-tutorials/00-Introduction`
        - :doc:`examples/00-Basic-tutorials/04-Scene-properties`
        - :doc:`examples/00-Basic-tutorials/05-Lighting-setups`
        - :doc:`examples/02-Advanced-topics/01-Devices`
    """

    def __init__(self, device=None, camera=None, lights=None):
        if device is None:
            device = Device()

        self._device = device
        self._scene = self.device.module.Scene(self.device._device)
        self.geometry = []
        if camera is None:
            self.camera = globals()['camera'].Orthographic(position=(0, 0, 100),
                                                           look_at=(0, 0, 0),
                                                           up=(0, 1, 0),
                                                           height=100)
        else:
            self.camera = camera

        if lights is None:
            self.lights = light.rembrandt()
        else:
            self.lights = lights

        self._tracer = None

    def get_extents(self):
        """Get the extents of the scene.

        Returns:
            (3,2) `numpy.ndarray` of ``numpy.float32``: The lower left and\
                upper right corners of the scene.
        """
        if len(self.geometry) == 0:
            return numpy.array([[0, 0, 0], [0, 0, 0]], dtype=numpy.float32)

        scene_extents = self.geometry[0].get_extents()
        for geom in self.geometry[1:]:
            extents = geom.get_extents()
            scene_extents[0, :] = numpy.min(
                [scene_extents[0, :], extents[0, :]], axis=0)
            scene_extents[1, :] = numpy.max(
                [scene_extents[1, :], extents[1, :]], axis=0)

        return scene_extents

    @property
    def device(self):
        """Device: Device this `Scene` is attached to."""
        return self._device

    @property
    def camera(self):
        """camera.Camera: Camera view parameters."""
        return camera._from_cpp(self._scene.getCamera())

    @camera.setter
    def camera(self, value):
        if isinstance(value, camera.Camera):
            self._scene.setCamera(value._camera)
        else:
            raise TypeError(f"camera {value} is not a fresnel.camera.Camera")

    @property
    def background_color(self):
        """((3, ) `numpy.ndarray` of ``numpy.float32``): Background color \
          linear RGB.

        Note:
            Use `fresnel.color.linear` to convert standard sRGB colors into the
            linear color space used by fresnel.
        """
        c = self._scene.getBackgroundColor()
        return numpy.array([c.r, c.g, c.b], dtype=numpy.float32)

    @background_color.setter
    def background_color(self, value):
        self._scene.setBackgroundColor(_common.RGBf(*value))

    @property
    def background_alpha(self):
        """float: Background alpha (opacity) in the range [0,1]."""
        return self._scene.getBackgroundAlpha()

    @background_alpha.setter
    def background_alpha(self, value):
        self._scene.setBackgroundAlpha(value)

    @property
    def lights(self):
        """list[Light]: Lights in the scene.

        `lights` is a sequence of up to 4 directional lights that apply to the
        scene. Each light has a direction, color, and size.
        """
        return light._LightListProxy(self._scene.getLights())

    @lights.setter
    def lights(self, values):
        tmp = light._LightListProxy()
        for v in values:
            tmp.append(v)

        self._scene.setLights(tmp._lights)


def preview(scene, w=600, h=370, anti_alias=True):
    """Preview a scene.

    Args:
        scene (`Scene`): Scene to render.
        w (int): Output image width (in pixels).
        h (int): Output image height (in pixels).
        anti_alias (bool): Whether to perform anti-aliasing.

    :py:func:`preview` is a shortcut that renders output with `tracer.Preview`.
    """
    t = tracer.Preview(scene.device, w=w, h=h, anti_alias=anti_alias)
    return t.render(scene)


def pathtrace(scene, w=600, h=370, samples=64, light_samples=1):
    """Path trace a scene.

    Args:
        scene (`Scene`): Scene to render.
        w (int): Output image width (in pixels).
        h (int): Output image height (in pixels).
        samples (int): Number of times to sample the pixels of the scene.

        light_samples (int): Number of light samples to take for each pixel
            sample.

    :py:func:`pathtrace` is a shortcut that renders output with `tracer.Path`.
    """
    t = tracer.Path(scene.device, w=w, h=h)
    t.sample(scene, samples=samples, light_samples=light_samples)
    return t.output



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test the Device class."""

import fresnel


def test_cpu():
    """Test the CPU device."""
    if 'cpu' in fresnel.Device.available_modes:
        fresnel.Device(mode='cpu')


def test_cpu_limit():
    """Test the cpu n argument."""
    if 'cpu' in fresnel.Device.available_modes:
        fresnel.Device(mode='cpu', n=2)


def test_gpu():
    """Test the GPU device."""
    if 'gpu' in fresnel.Device.available_modes:
        fresnel.Device(mode='gpu')


def test_gpu_limit():
    """Test the GPU n argument."""
    if 'gpu' in fresnel.Device.available_modes:
        fresnel.Device(mode='gpu', n=1)



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test the Scene class."""

import fresnel
import numpy
from collections import namedtuple
import PIL
import conftest
import os
import pathlib

dir_path = pathlib.Path(os.path.realpath(__file__)).parent


def test_background_color(device_):
    """Test the background_color property."""
    scene = fresnel.Scene(device=device_)

    scene.background_color = fresnel.color.linear((0.125, 0.75, 0.375))

    numpy.testing.assert_array_equal(scene.background_color,
                                     fresnel.color.linear((0.125, 0.75, 0.375)))

    scene.background_alpha = 0.5

    assert scene.background_alpha == 0.5

    scene.camera = fresnel.camera.Orthographic(position=(0, 0, 10),
                                               look_at=(0, 0, 0),
                                               up=(0, 1, 0),
                                               height=7)
    buf_proxy = fresnel.preview(scene, w=100, h=100, anti_alias=False)
    buf = buf_proxy[:]

    numpy.testing.assert_array_equal(
        buf[:, :, 3],
        numpy.ones(shape=(100, 100), dtype=buf.dtype) * 128)
    numpy.testing.assert_array_equal(
        buf[:, :, 0:3],
        numpy.ones(shape=(100, 100, 3), dtype=buf.dtype) * (32, 191, 96))


def test_camera(scene_hex_sphere_, generate=False):
    """Test the camera property."""
    scene_hex_sphere_.camera = fresnel.camera.Orthographic(position=(1, 0, 10),
                                                           look_at=(1, 0, 0),
                                                           up=(0, 1, 0),
                                                           height=6)

    numpy.testing.assert_array_equal(scene_hex_sphere_.camera.position,
                                     (1, 0, 10))
    numpy.testing.assert_array_equal(scene_hex_sphere_.camera.look_at,
                                     (1, 0, 0))
    numpy.testing.assert_array_equal(scene_hex_sphere_.camera.up, (0, 1, 0))
    assert scene_hex_sphere_.camera.height == 6

    buf_proxy = fresnel.preview(scene_hex_sphere_,
                                w=100,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_scene.test_camera.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:], dir_path / 'reference' / 'test_scene.test_camera.png')


def test_light_dir(scene_hex_sphere_, generate=False):
    """Test the lights property."""
    scene_hex_sphere_.lights[0].direction = (1, 0, 0)
    scene_hex_sphere_.lights[0].color = (0.5, 0.5, 0.5)
    assert scene_hex_sphere_.lights[0].direction == (1, 0, 0)
    assert scene_hex_sphere_.lights[0].color == (0.5, 0.5, 0.5)

    buf_proxy = fresnel.preview(scene_hex_sphere_,
                                w=100,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_scene.test_light_dir.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_scene.test_light_dir.png')


def test_multiple_geometries(device_, generate=False):
    """Test multiple geometries."""
    scene = fresnel.Scene(lights=conftest.test_lights())
    scene.camera = fresnel.camera.Orthographic(position=(0, 0, 10),
                                               look_at=(0, 0, 0),
                                               up=(0, 1, 0),
                                               height=7)

    geom1 = fresnel.geometry.Sphere(scene,
                                    position=[[-4, 1, 0], [-4, -1, 0],
                                              [-2, 1, 0], [-2, -1, 0]],
                                    radius=1.0)
    geom1.material = fresnel.material.Material(solid=1.0,
                                               color=fresnel.color.linear(
                                                   [0.42, 0.267, 1]))
    geom1.outline_width = 0.12

    geom2 = fresnel.geometry.Sphere(scene,
                                    position=[[4, 1, 0], [4, -1, 0], [2, 1, 0],
                                              [2, -1, 0]],
                                    radius=1.0)
    geom2.material = fresnel.material.Material(solid=0.0,
                                               color=fresnel.color.linear(
                                                   [1, 0.874, 0.169]))

    buf_proxy = fresnel.preview(scene, w=200, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_scene.test_multiple_geometries1.png', 'wb'),
            'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_scene.test_multiple_geometries1.png')

    geom1.disable()

    buf_proxy = fresnel.preview(scene, w=200, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_scene.test_multiple_geometries2.png', 'wb'),
            'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_scene.test_multiple_geometries2.png')

    geom1.enable()

    buf_proxy = fresnel.preview(scene, w=200, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_scene.test_multiple_geometries3.png', 'wb'),
            'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_scene.test_multiple_geometries3.png')

    geom2.remove()

    buf_proxy = fresnel.preview(scene, w=200, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_scene.test_multiple_geometries4.png', 'wb'),
            'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_scene.test_multiple_geometries4.png')


if __name__ == '__main__':
    struct = namedtuple("struct", "param")
    device = conftest.device(struct(('cpu', None)))

    scene = conftest.scene_hex_sphere(device)
    test_camera(scene, generate=True)

    scene = conftest.scene_hex_sphere(device)
    test_light_dir(scene, generate=True)

    scene = conftest.scene_hex_sphere(device)
    test_multiple_geometries(scene, generate=True)



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Ray tracers process a `Scene` and render output images.

* `Preview` generates a quick approximate render.
* `Path` which provides soft shadows, reflections, and other effects.

See Also:
    Tutorials:

    - :doc:`examples/00-Basic-tutorials/00-Introduction`
    - :doc:`examples/02-Advanced-topics/02-Tracer-methods`
"""

import numpy
from . import util
from . import _common


class Tracer(object):
    """Base class for all ray tracers.

    `Tracer` provides operations common to all ray tracer classes.

    Each `Tracer` instance stores a pixel output buffer. When you `render` a
    `Scene`, the `output` is updated.

    Note:
        You cannot instantiate `Tracer` directly. Use one of the subclasses.
    """

    def __init__(self):
        raise RuntimeError("Use a specific tracer class")

    def resize(self, w, h):
        """Resize the output buffer.

        Args:
            w (int): New output buffer width.
            h (int): New output buffer height.

        Warning:
            `resize` clears the output buffer.
        """
        self._tracer.resize(w, h)

    def render(self, scene):
        """Render a scene.

        Args:
            scene (`Scene <fresnel.Scene>`): The scene to render.

        Returns:
            A reference to the current output buffer as a
            `fresnel.util.ImageArray`.

        Render the given scene and write the resulting pixels into the output
        buffer.
        """
        self._tracer.render(scene._scene)
        return self.output

    def enable_highlight_warning(self, color=(1, 0, 1)):
        """Enable highlight clipping warnings.

        When a pixel in the rendered image is too bright to represent, make that
        pixel the given *color* to flag the problem to the user.

        Args:
            color (tuple): Color to make the highlight warnings.
        """
        self._tracer.enableHighlightWarning(_common.RGBf(*color))

    def disable_highlight_warning(self):
        """Disable the highlight clipping warnings."""
        self._tracer.disableHighlightWarning()

    def histogram(self):
        """Compute a histogram of the image.

        The histogram is computed as a lightness in the sRGB color space. The
        histogram is computed only over the visible pixels in the image, fully
        transparent pixels are ignored. The returned histogram is nbins x 4, the
        first column contains the lightness histogram and the next 3 contain
        R,B, and G channel histograms respectively.

        Return:
            (histogram, bin_positions).
        """
        a = numpy.array(self.linear_output[:])
        img_sel = a[:, :, 3] > 0
        img = a[img_sel]
        r = img[:, 0]
        g = img[:, 1]
        b = img[:, 2]
        l = 0.21 * r + 0.72 * g + 0.07 * b  # noqa: E741 - allow l as a name
        gamma_l = l**(1 / 2.2)

        n = 512
        l_hist, bins = numpy.histogram(gamma_l, bins=n, range=[0, 1])
        r_hist, bins = numpy.histogram(r**(1 / 2.2), bins=n, range=[0, 1])
        g_hist, bins = numpy.histogram(g**(1 / 2.2), bins=n, range=[0, 1])
        b_hist, bins = numpy.histogram(b**(1 / 2.2), bins=n, range=[0, 1])

        out = numpy.stack((l_hist, r_hist, g_hist, b_hist), axis=1)

        return out, bins[1:]

    @property
    def output(self):
        """ImageArray: Reference to the current output buffer.

        Note:
            The output buffer is modified by `render` and `resize`.
        """
        return util.ImageArray(self._tracer.getSRGBOutputBuffer(), geom=None)

    @property
    def linear_output(self):
        """Array: Reference to the current output buffer in linear color space.

        Note:
            The output buffer is modified by `render` and `resize`.
        """
        return util.Array(self._tracer.getLinearOutputBuffer(), geom=None)

    @property
    def seed(self):
        """int: Random number seed."""
        return self._tracer.getSeed()

    @seed.setter
    def seed(self, value):
        self._tracer.setSeed(value)


class Preview(Tracer):
    """Preview ray tracer.

    Args:
        device (`Device`): Device to use.

        w (int): Output image width.

        h (int): Output image height.

        anti_alias (bool): Whether to perform anti-aliasing. If True, uses an
            64 samples.

    .. rubric:: Overview

    The `Preview` tracer produces a preview of the scene quickly. It
    approximates the effect of light on materials. The output of the `Preview`
    tracer will look very similar to that from the `Path` tracer, but will miss
    soft shadows, reflection, transmittance, depth of field and other effects.

    .. rubric:: Anti-aliasing

    The default value of `anti_alias` is True to smooth sharp edges in
    the image. The anti-aliasing level corresponds to ``aa_level=3`` in fresnel
    versions up to 0.11.0. Different `seed` values will result in different
    output images.
    """

    def __init__(self, device, w, h, anti_alias=True):
        self.device = device
        self._tracer = device.module.TracerDirect(device._device, w, h)
        self.anti_alias = anti_alias

    @property
    def anti_alias(self):
        """bool: Whether to perform anti-aliasing."""
        return self._tracer.getAntialiasingN() > 1

    @anti_alias.setter
    def anti_alias(self, value):
        if value:
            self._tracer.setAntialiasingN(8)
        else:
            self._tracer.setAntialiasingN(1)


class Path(Tracer):
    """Path tracer.

    Args:
        device (`Device`): Device to use.
        w (int): Output image width.
        h (int): Output image height.

    The path tracer applies advanced lighting effects, including soft shadows,
    reflections, and depth of field. It operates by Monte Carlo sampling. Each
    call to `render` performs one sample per pixel. The `output` image is the
    mean of all the samples. Many samples are required to produce a smooth
    image. `sample` provides a convenience API to make many samples with a
    single call.
    """

    def __init__(self, device, w, h):
        self.device = device
        self._tracer = device.module.TracerPath(device._device, w, h, 1)

    def reset(self):
        """Clear the output buffer.

        Start sampling a new image. Increment the random number seed so that the
        new image is statistically independent from the previous.
        """
        self._tracer.reset()

    def sample(self, scene, samples, reset=True, light_samples=1):
        r"""Sample the image.

        Args:
            scene (`Scene`): The scene to render.

            samples (int): The number of samples to take per pixel.

            reset (bool): When True, call `reset` before sampling

            light_samples (int): The number of light samples per primary camera
                ray.

        As an unbiased renderer, the sampling noise will scale as
        :math:`\frac{1}{\sqrt{\text{total_samples}}}`, where ``total_samples``
        is ``samples*light_samples``.

        The ``samples`` parameter controls the number of samples from the camera
        (depth of field and antialiasing). ``light_samples`` is the number of
        rays shot from the first intersection of the primary camera ray.

        Using ``(samples=N, light_samples=1)`` would have an equal number of
        camera and lighting samples and would produce an excellent image. Using
        ``(samples=N // M, light_samples=M)`` (where ``M`` is some integer
        10-100) will similarly produce a similarly good image, provided ``N >>
        M``, and may have better performance. On the GPU, using ``light_samples
        > 1`` can boost performance moderately. On the CPU, it can boost
        performance slightly due to improved cache coherency.

        Returns:
            ImageArray: A reference to the current `output` buffer.

        Note:
            When *reset* is ``False``, subsequent calls to `sample` will
            continue to add samples to the current output image. Use the same
            number of light samples when sampling an image in this way.
        """
        if reset:
            self.reset()

        self._tracer.setLightSamples(light_samples)

        for i in range(samples):
            out = self.render(scene)

        # reset the number of light samples to 1 to avoid side effects with
        # future calls to render() by the user
        self._tracer.setLightSamples(1)

        return out



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test the Perspective camera."""

import fresnel
import numpy
from collections import namedtuple
import conftest
import pytest


def scene(device):
    """Create a test scene."""
    scene = fresnel.Scene(device, lights=conftest.test_lights())

    position = []
    for i in range(-8, 8):
        for j in range(-8, 8):
            position.append([i, 0, j])

    mat = fresnel.material.Material(
        color=fresnel.color.linear([0.42, 0.267, 0.9]))
    fresnel.geometry.Sphere(
        scene,
        position=position,
        radius=0.7,
        material=mat,
    )

    scene.camera = fresnel.camera.Perspective(position=(0, 4, -10),
                                              look_at=(0, 0, 0),
                                              up=(0, 1, 0),
                                              focus_distance=10)

    return scene


@pytest.fixture(scope='function')
def scene_(device_):
    """Pytest fixture to create a test scene."""
    return scene(device_)


def test_perspective_attributes():
    """Test all properties."""
    cam = fresnel.camera.Perspective(position=(0, 0, -10),
                                     look_at=(0, 0, 0),
                                     up=(0, 1, 0),
                                     height=0.25,
                                     focal_length=1.25,
                                     focus_distance=20,
                                     f_stop=1.5)
    numpy.testing.assert_array_equal(cam.position, (0, 0, -10))
    numpy.testing.assert_array_equal(cam.look_at, (0, 0, 0))
    numpy.testing.assert_array_equal(cam.up, (0, 1, 0))
    assert cam.height == 0.25
    assert cam.focal_length == 1.25
    assert cam.focus_distance == 20
    assert cam.f_stop == 1.5

    cam2 = eval(repr(cam))
    numpy.testing.assert_array_equal(cam2.position, (0, 0, -10))
    numpy.testing.assert_array_equal(cam2.look_at, (0, 0, 0))
    numpy.testing.assert_array_equal(cam2.up, (0, 1, 0))
    assert cam2.height == 0.25
    assert cam2.focal_length == 1.25
    assert cam2.focus_distance == 20
    assert cam2.f_stop == 1.5

    numpy.testing.assert_allclose(cam.depth_of_field, 0.225007119365887)
    cam.depth_of_field = 1
    numpy.testing.assert_allclose(cam.depth_of_field, 1)
    numpy.testing.assert_allclose(cam.f_stop, 6.662505149841309)

    numpy.testing.assert_allclose(cam.focus_on, [0, 0, 10])
    cam.focus_on = [20, -50, 5]
    numpy.testing.assert_allclose(cam.focus_on, [0, 0, 5])
    numpy.testing.assert_allclose(cam.focus_distance, 15)

    numpy.testing.assert_allclose(cam.vertical_field_of_view,
                                  0.19933730498232408)
    cam.vertical_field_of_view = 0.4
    numpy.testing.assert_allclose(cam.focal_length, 0.6166443824768066)


def test_render(scene_, generate=False):
    """Test that preview can render a scene."""
    conftest.check_preview_render(scene_,
                                  "test_camera_perspective.test_render",
                                  generate=generate)


def test_focal_length(scene_, generate=False):
    """Test the focal_length property."""
    scene_.camera.focal_length = 0.25

    conftest.check_preview_render(scene_,
                                  "test_camera_perspective.test_focal_length",
                                  generate=generate)


def test_height(scene_, generate=False):
    """Test the height property."""
    scene_.camera.height = .125

    conftest.check_preview_render(scene_,
                                  "test_camera_perspective.test_height",
                                  generate=generate)


def test_pathtrace(scene_, generate=False):
    """Test that Preview works with pathtrace."""
    conftest.check_pathtrace_render(scene_,
                                    "test_camera_perspective.test_pathtrace",
                                    generate=generate)


def test_f_stop(scene_, generate=False):
    """Test the f_stop property."""
    scene_.camera.f_stop = 0.2

    conftest.check_pathtrace_render(scene_,
                                    "test_camera_perspective.test_f_stop",
                                    generate=generate)


def test_focus_distance(scene_, generate=False):
    """Test the focus_distance property."""
    scene_.camera.f_stop = 0.2
    scene_.camera.focus_distance = 6

    conftest.check_pathtrace_render(
        scene_,
        "test_camera_perspective.test_focal_distance",
        generate=generate)


if __name__ == '__main__':
    struct = namedtuple("struct", "param")
    device = conftest.device(struct(('cpu', None)))

    for method in [
            test_render,
            test_focal_length,
            test_height,
            test_pathtrace,
            test_f_stop,
            test_focus_distance,
    ]:
        s = scene(device)
        method(s, generate=True)



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test the Cylinder geometry."""

import fresnel
import numpy
from collections import namedtuple
import PIL
import conftest
import pytest
import os
import pathlib

dir_path = pathlib.Path(os.path.realpath(__file__)).parent


def scene_four_cylinders(device):
    """Create a test scene with four cylinders."""
    scene = fresnel.Scene(device, lights=conftest.test_lights())

    position = [[[-5, -5, 0], [-5, 5, 0]], [[5, -5, -5], [5, 5, 5]],
                [[3, 3, -3], [-3, -3, -3]], [[-2, 2, 2], [2, -2, -2]]]

    fresnel.geometry.Cylinder(
        scene,
        points=position,
        radius=1.0,
        color=[0.9, 0.9, 0.9],
        material=fresnel.material.Material(
            color=fresnel.color.linear([0.42, 0.267, 1])),
    )

    scene.camera = fresnel.camera.Orthographic(position=(0, 2, 10),
                                               look_at=(0, 0, 0),
                                               up=(0, 1, 0),
                                               height=15)

    return scene


@pytest.fixture(scope='function')
def scene_four_cylinders_(device_):
    """Pytest fixture to create a test scene."""
    return scene_four_cylinders(device_)


def test_render(scene_four_cylinders_, generate=False):
    """Test that Cylinder renders properly."""
    buf_proxy = fresnel.preview(scene_four_cylinders_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_clyinder.test_render.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_clyinder.test_render.png')


def test_radius(scene_four_cylinders_, generate=False):
    """Test the radius property."""
    geometry = scene_four_cylinders_.geometry[0]

    r = numpy.array([0.5, 0.6, 0.8, 1.0], dtype=numpy.float32)
    geometry.radius[:] = r
    numpy.testing.assert_array_equal(r, geometry.radius[:])

    buf_proxy = fresnel.preview(scene_four_cylinders_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_clyinder.test_radius.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_clyinder.test_radius.png')


def test_points(scene_four_cylinders_, generate=False):
    """Test the points property."""
    geometry = scene_four_cylinders_.geometry[0]

    p = numpy.array([[[-5, -5, 0], [-5, 5, 0]], [[-3, 5, 0], [3, 5, 0]],
                     [[5, 5, 0], [5, -5, -0]], [[3, -5, 0], [-3, -5, 0]]],
                    dtype=numpy.float32)
    geometry.points[:] = p
    numpy.testing.assert_array_equal(p, geometry.points[:])

    buf_proxy = fresnel.preview(scene_four_cylinders_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_clyinder.test_position.png', 'wb'),
            'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_clyinder.test_position.png')


def test_color(scene_four_cylinders_, generate=False):
    """Test the color property."""
    geometry = scene_four_cylinders_.geometry[0]
    geometry.material.primitive_color_mix = 1.0

    c = numpy.array(
        [[[0.9, 0, 0], [0, 0.9, 0]], [[0, 0, 0.9], [0, 0.9, 0.9]],
         [[0.9, 0.9, 0], [0.9, 0, 0.9]], [[0.1, 0.2, 0.9], [0.9, 0.2, 0.3]]],
        dtype=numpy.float32)
    geometry.color[:] = c
    numpy.testing.assert_array_equal(c, geometry.color[:])

    buf_proxy = fresnel.preview(scene_four_cylinders_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_clyinder.test_color.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_clyinder.test_color.png')


def test_outline(scene_four_cylinders_, generate=False):
    """Test that outlines render properly."""
    geometry = scene_four_cylinders_.geometry[0]
    geometry.outline_width = 0.3

    buf_proxy = fresnel.preview(scene_four_cylinders_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_clyinder.test_outline.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_clyinder.test_outline.png')


if __name__ == '__main__':
    struct = namedtuple("struct", "param")
    device = conftest.device(struct(('cpu', None)))

    scene = scene_four_cylinders(device)
    test_render(scene, generate=True)

    scene = scene_four_cylinders(device)
    test_radius(scene, generate=True)

    scene = scene_four_cylinders(device)
    test_points(scene, generate=True)

    scene = scene_four_cylinders(device)
    test_color(scene, generate=True)

    scene = scene_four_cylinders(device)
    test_outline(scene, generate=True)



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test the ConvexPolyhedron geometry."""

import fresnel
import numpy
from collections import namedtuple
import PIL
import conftest
import pytest
import math
import itertools
import os
import pathlib

dir_path = pathlib.Path(os.path.realpath(__file__)).parent


def scene_eight_polyhedra(device):
    """Create a test scene with eight polyhedra."""
    scene = fresnel.Scene(device, lights=conftest.test_lights())

    # place eight polyhedra
    position = []
    for k in range(2):
        for i in range(2):
            for j in range(2):
                position.append([2.5 * i, 2.5 * j, 2.5 * k])

    # create the polyhedron faces
    origins = []
    normals = []
    colors = []

    for v in [-1, 1]:
        origins.append([v, 0, 0])
        normals.append([v, 0, 0])
        origins.append([0, v, 0])
        normals.append([0, v, 0])
        origins.append([0, 0, v])
        normals.append([0, 0, v])
        colors.append([178 / 255, 223 / 255, 138 / 255])
        colors.append([178 / 255, 223 / 255, 138 / 255])
        colors.append([178 / 255, 223 / 255, 138 / 255])

    for x in [-1, 1]:
        for y in [-1, 1]:
            for z in [-1, 1]:
                normals.append([x, y, z])
                origins.append([x * 0.75, y * 0.75, z * 0.75])
                colors.append([166 / 255, 206 / 255, 227 / 255])

    poly_info = {
        'face_normal': normals,
        'face_origin': origins,
        'radius': math.sqrt(3),
        'face_color': fresnel.color.linear(colors)
    }
    geometry = fresnel.geometry.ConvexPolyhedron(scene,
                                                 poly_info,
                                                 position=position)

    geometry.material = \
        fresnel.material.Material(color=fresnel.color.linear([1.0, 0, 0]),
                                  roughness=0.8,
                                  specular=0.5,
                                  primitive_color_mix=0.0)
    geometry.orientation[:] = [1, 0, 0, 0]

    scene.camera = fresnel.camera.Orthographic(position=(20, 20, 20),
                                               look_at=(0, 0, 0),
                                               up=(0, 1, 0),
                                               height=7)

    return scene


@pytest.fixture(scope='function')
def scene_eight_polyhedra_(device_):
    """Pytest fixture to create a test scene."""
    return scene_eight_polyhedra(device_)


def test_render(scene_eight_polyhedra_, generate=False):
    """Test that convex polyhedra render properly."""
    buf_proxy = fresnel.preview(scene_eight_polyhedra_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_convex_polyhedron.test_render.png',
                 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:], dir_path / 'reference'
            / 'test_geometry_convex_polyhedron.test_render.png')


def test_outline(scene_eight_polyhedra_, generate=False):
    """Test that face outlines render properly."""
    geometry = scene_eight_polyhedra_.geometry[0]
    geometry.outline_width = 0.1

    buf_proxy = fresnel.preview(scene_eight_polyhedra_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_convex_polyhedron.test_outline.png',
                 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:], dir_path / 'reference'
            / 'test_geometry_convex_polyhedron.test_outline.png')


def test_face_color(scene_eight_polyhedra_, generate=False):
    """Test that faces can be colored individually."""
    buf_proxy = fresnel.preview(scene_eight_polyhedra_,
                                w=150,
                                h=100,
                                anti_alias=False)

    geometry = scene_eight_polyhedra_.geometry[0]
    geometry.color_by_face = 1.0
    geometry.material.primitive_color_mix = 1.0

    buf_proxy = fresnel.preview(scene_eight_polyhedra_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_convex_polyhedron.test_face_color.png',
                 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:], dir_path / 'reference'
            / 'test_geometry_convex_polyhedron.test_face_color.png')


def test_convert_cube():
    """Sanity checks on converting vertices to origins and normals."""
    pms = [+1, -1]
    cube_verts = numpy.array([x for x in itertools.product(pms, repeat=3)])

    poly_info = fresnel.util.convex_polyhedron_from_vertices(cube_verts)
    assert poly_info['face_origin'].shape[0] == 6
    assert poly_info['face_normal'].shape[0] == 6
    for f in poly_info['face_sides']:
        assert f == 4  # should all be squares
    assert poly_info['radius'] == numpy.sqrt(3)


def test_face_merge_cube():
    """Add a point into the middle and make sure no new faces are created."""
    pms = [+1, -1]
    cube_verts = numpy.array([x for x in itertools.product(pms, repeat=3)])

    cube_verts = numpy.concatenate((cube_verts, [[0.5, 0.5, 1.0]]))
    poly_info = fresnel.util.convex_polyhedron_from_vertices(cube_verts)
    assert poly_info['face_origin'].shape[0] == 6
    assert poly_info['face_normal'].shape[0] == 6
    for f in poly_info['face_sides']:
        assert f == 4  # should all be squares
    assert poly_info['radius'] == numpy.sqrt(3)


if __name__ == '__main__':
    struct = namedtuple("struct", "param")
    device = conftest.device(struct(('cpu', None)))

    scene = scene_eight_polyhedra(device)
    test_render(scene, generate=True)

    scene = scene_eight_polyhedra(device)
    test_outline(scene, generate=True)

    scene = scene_eight_polyhedra(device)
    test_face_color(scene, generate=True)



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test the Direct tracer."""

import fresnel
from collections import namedtuple
import PIL
import conftest
import os
import pathlib

dir_path = pathlib.Path(os.path.realpath(__file__)).parent


def test_render(scene_hex_sphere_, generate=False):
    """Test rendering a scene with 6 spheres."""
    tracer = fresnel.tracer.Preview(device=scene_hex_sphere_.device,
                                    w=100,
                                    h=100,
                                    anti_alias=False)
    buf = tracer.output[:]
    assert buf.shape == (100, 100, 4)

    buf_proxy = tracer.render(scene_hex_sphere_)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_tracer_direct.test_render.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_tracer_direct.test_render.png')


def test_render_aa(scene_hex_sphere_, generate=False):
    """Test rendering with anti-aliasing."""
    tracer = fresnel.tracer.Preview(device=scene_hex_sphere_.device,
                                    w=100,
                                    h=100,
                                    anti_alias=True)
    buf = tracer.output[:]
    assert buf.shape == (100, 100, 4)

    buf_proxy = tracer.render(scene_hex_sphere_)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_tracer_direct.test_render_aa.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_tracer_direct.test_render_aa.png')


def test_resize(scene_hex_sphere_, generate=False):
    """Test the resize() method."""
    tracer = fresnel.tracer.Preview(device=scene_hex_sphere_.device,
                                    w=100,
                                    h=100,
                                    anti_alias=False)
    buf = tracer.output[:]
    assert buf.shape == (100, 100, 4)

    tracer.resize(w=200, h=300)
    buf = tracer.output[:]
    assert buf.shape == (300, 200, 4)


if __name__ == '__main__':
    struct = namedtuple("struct", "param")
    device = conftest.device(struct(('cpu', None)))

    scene = conftest.scene_hex_sphere(device)
    test_render(scene, generate=True)

    scene = conftest.scene_hex_sphere(device)
    test_render_aa(scene, generate=True)



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test the Polygon geometry."""

import fresnel
import numpy
from collections import namedtuple
import PIL
import pytest
import conftest
import os
import pathlib

dir_path = pathlib.Path(os.path.realpath(__file__)).parent


def scene_rounded_polygons(device):
    """Create a test scene with polygons."""
    scene = fresnel.Scene(device, lights=conftest.test_lights())

    mat = fresnel.material.Material(color=fresnel.color.linear([0.42, 0.267,
                                                                1]),
                                    solid=1)

    fresnel.geometry.Polygon(scene,
                             N=2,
                             rounding_radius=0.3,
                             vertices=[[-1, -1], [1, -1], [1, 1], [0, 0],
                                       [-1, 1]],
                             position=[[-1.5, 0], [1.5, 0]],
                             angle=[0.1, -0.2],
                             color=[[0, 0, 1], [0, 1, 0]],
                             material=mat)

    scene.camera = fresnel.camera.Orthographic(position=(0, 0, -2),
                                               look_at=(0, 0, 0),
                                               up=(0, 1, 0),
                                               height=5)

    return scene


@pytest.fixture(scope='function')
def scene_rounded_polygons_(device_):
    """Pytest fixture to create test scene."""
    return scene_rounded_polygons(device_)


def scene_polygons(device):
    """Create a test scene with polygons."""
    scene = fresnel.Scene(device, lights=conftest.test_lights())

    mat = fresnel.material.Material(color=fresnel.color.linear([0.42, 0.267,
                                                                1]),
                                    solid=1)

    fresnel.geometry.Polygon(scene,
                             N=2,
                             rounding_radius=0,
                             vertices=[[-1, -1], [1, -1], [1, 1], [0, 0],
                                       [-1, 1]],
                             position=[[-1.5, 0], [1.5, 0]],
                             angle=[0.1, -0.2],
                             color=[[0, 0, 1], [0, 1, 0]],
                             material=mat)

    scene.camera = fresnel.camera.Orthographic(position=(0, 0, -2),
                                               look_at=(0, 0, 0),
                                               up=(0, 1, 0),
                                               height=5)

    return scene


@pytest.fixture(scope='function')
def scene_polygons_(device_):
    """Pytest fixture to create test scene."""
    return scene_polygons(device_)


def test_render(scene_polygons_, generate=False):
    """Test that Polygons render properly."""
    buf_proxy = fresnel.preview(scene_polygons_, w=150, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_polygon.test_render.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_polygon.test_render.png')


def test_rounded(scene_rounded_polygons_, generate=False):
    """Test that rounded polygons render properly."""
    geometry = scene_rounded_polygons_.geometry[0]
    geometry.outline_width = 0.1

    buf_proxy = fresnel.preview(scene_rounded_polygons_,
                                w=150,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_polygon.test_rounded.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_polygon.test_rounded.png')


def test_angle(scene_polygons_, generate=False):
    """Test that polygons can be rotated."""
    geometry = scene_polygons_.geometry[0]

    a = numpy.array([-0.8, 0.5], dtype=numpy.float32)
    geometry.angle[:] = a
    numpy.testing.assert_array_equal(a, geometry.angle[:])

    buf_proxy = fresnel.preview(scene_polygons_, w=150, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_polygon.test_angle.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_polygon.test_angle.png')


def test_position(scene_polygons_, generate=False):
    """Test the position property."""
    geometry = scene_polygons_.geometry[0]

    p = numpy.array([[-2, 0], [3, 0]], dtype=numpy.float32)
    geometry.position[:] = p
    numpy.testing.assert_array_equal(p, geometry.position[:])

    buf_proxy = fresnel.preview(scene_polygons_, w=150, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_polygon.test_position.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_polygon.test_position.png')


def test_color(scene_polygons_, generate=False):
    """Test the color property."""
    geometry = scene_polygons_.geometry[0]
    geometry.material.primitive_color_mix = 1.0

    c = numpy.array(
        [fresnel.color.linear([0, 1, 0]),
         fresnel.color.linear([1, 0, 0])],
        dtype=numpy.float32)
    geometry.color[:] = c
    numpy.testing.assert_array_equal(c, geometry.color[:])

    buf_proxy = fresnel.preview(scene_polygons_, w=150, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_polygon.test_color.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_polygon.test_color.png')


def test_outline(scene_polygons_, generate=False):
    """Test that outlines render properly."""
    geometry = scene_polygons_.geometry[0]
    geometry.outline_width = 0.1

    buf_proxy = fresnel.preview(scene_polygons_, w=150, h=100, anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_polygon.test_outline.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_polygon.test_outline.png')


if __name__ == '__main__':
    struct = namedtuple("struct", "param")
    device = conftest.device(struct(('cpu', None)))

    scene = scene_polygons(device)
    test_render(scene, generate=True)

    scene = scene_rounded_polygons(device)
    test_rounded(scene, generate=True)

    scene = scene_polygons(device)
    test_angle(scene, generate=True)

    scene = scene_polygons(device)
    test_position(scene, generate=True)

    scene = scene_polygons(device)
    test_color(scene, generate=True)

    scene = scene_polygons(device)
    test_outline(scene, generate=True)



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Geometric primitives.

Geometry defines objects that are visible in a `Scene`. The base class
`Geometry` provides common operations and properties. Instantiate specific
geometry class to add objects to a `Scene`.

See Also:
    Tutorials:

    - :doc:`examples/00-Basic-tutorials/01-Primitive-properties`
    - :doc:`examples/00-Basic-tutorials/02-Material-properties`
    - :doc:`examples/00-Basic-tutorials/03-Outline-materials`
    - :doc:`examples/02-Advanced-topics/00-Multiple-geometries`
"""

from . import material
from . import util
import numpy


class Geometry(object):
    """Geometry base class.

    `Geometry` provides operations and properties common to all geometry
    classes.

    Note:
        You cannot instantiate a Geometry directly. Use one of the subclasses.
    """

    def __init__(self):
        raise RuntimeError("Use a specific geometry class")

    def enable(self):
        """Enable the geometry.

        When enabled, the geometry will be visible in the `Scene`.

        See Also:
            `disable`
        """
        self._geometry.enable()

    def disable(self):
        """Disable the geometry.

        When disabled, the geometry will not visible in the `Scene`.

        See Also:
            `enable`
        """
        self._geometry.disable()

    def remove(self):
        """Remove the geometry from the scene.

        After calling `remove`, the geometry is no longer part of the scene. It
        cannot be added back into the scene. Use `disable` and `enable` hide
        geometry reversibly.
        """
        self._geometry.remove()
        self.scene.geometry.remove(self)

    @property
    def material(self):
        """material.Material: Define how light interacts with the geometry."""
        return material._MaterialProxy(self)

    @material.setter
    def material(self, mat):
        self._geometry.setMaterial(mat._get_cpp_material())

    @property
    def outline_material(self):
        """material.Material: Define how light interacts with the geometry's \
            outline."""
        return material._OutlineMaterialProxy(self)

    @outline_material.setter
    def outline_material(self, mat):
        self._geometry.setOutlineMaterial(mat._get_cpp_material())

    @property
    def outline_width(self):
        """float: Width of the outline in scene units."""
        return self._geometry.getOutlineWidth()

    @outline_width.setter
    def outline_width(self, width):
        self._geometry.setOutlineWidth(width)


class Cylinder(Geometry):
    """Cylinder geometry.

    Define a set of spherocylinder primitives with individual start and end
    positions, radii, and colors.

    Args:
        scene (Scene): Add the geometry to this scene.

        points ((N, 2, 3) `numpy.ndarray` of ``float32``): *N* cylinder start
            and end points.

        radius ((N, ) `numpy.ndarray` of ``float32``): Radius of each cylinder.

        color ((N, 2, 3) `numpy.ndarray` of ``float32``): Color of each start
            and end point.

        N (int): Number of cylinders in the geometry. When `None`, determine
            *N* from *points*.

        material (material.Material): Define how light interacts with the
            geometry. When `None`, defaults to ``material.Material()``.

        outline_material (material.Material): Define how light interacts with \
            the geometry's outline. When `None`, defaults to solid black
            ``material.Material(solid=1, color=(0,0,0))``.

        outline_width (float): Width of the outline in scene units.

    See Also:
        Tutorials:

        - :doc:`examples/01-Primitives/01-Cylinder-geometry`

    Hint:
        Avoid costly memory allocations and type conversions by specifying
        primitive properties in the appropriate array type.

    Tip:
        When all cylinders are the same size or color, pass a single value
        and NumPy will broadcast it to all elements of the array.
    """

    def __init__(self,
                 scene,
                 points=((0, 0, 0), (0, 0, 0)),
                 radius=0.5,
                 color=(0, 0, 0),
                 N=None,
                 material=None,
                 outline_material=None,
                 outline_width=0.0):
        if N is None:
            N = len(points)

        self._geometry = scene.device.module.GeometryCylinder(scene._scene, N)
        if material is None:
            self.material = globals()['material'].Material()
        else:
            self.material = material

        if outline_material is None:
            self.outline_material = globals()['material'].Material(solid=1,
                                                                   color=(0, 0,
                                                                          0))
        else:
            self.outline_material = outline_material

        self.outline_width = outline_width

        self.points[:] = points
        self.radius[:] = radius
        self.color[:] = color

        self.scene = scene
        self.scene.geometry.append(self)

    def get_extents(self):
        """Get the extents of the geometry.

        Returns:
            (3,2) `numpy.ndarray` of ``float32``: The lower left and\
                upper right corners of the scene.
        """
        A = self.points[:, 0]
        B = self.points[:, 1]
        r = self.radius[:]
        r = r.reshape(len(r), 1)
        res = numpy.array([
            numpy.min([numpy.min(A - r, axis=0),
                       numpy.min(B - r, axis=0)],
                      axis=0),
            numpy.max([numpy.max(A + r, axis=0),
                       numpy.max(B + r, axis=0)],
                      axis=0)
        ])
        return res

    @property
    def points(self):
        """(N, 2, 3) `Array`: The start and end points of the cylinders."""
        return util.Array(self._geometry.getPointsBuffer(), geom=self)

    @property
    def radius(self):
        """(N, ) `Array`: The radii of the cylinders."""
        return util.Array(self._geometry.getRadiusBuffer(), geom=self)

    @property
    def color(self):
        """(N, 2, 3) `Array`: Color of each start and end point."""
        return util.Array(self._geometry.getColorBuffer(), geom=self)


class Box(Cylinder):
    """Box geometry.

    Generate a triclinic box outline with `spherocylinders <Cylinder>`.
    The geometry's material defaults to ``material.Material(solid=1.0)``.

    Args:
        scene (Scene): Add the geometry to this scene.

        box ((1, ), (3, ), or (6, ) `numpy.ndarray` of ``float32``): Box
            parameters.

        box_radius (float): Radius of box edges.

        box_color ((3, ) `numpy.ndarray` of ``float32``): Color of the box
            edges.

    Note:
        A 1-element *box* array expands to a cube. A 3-element *box* array
        ``[Lx, Ly, Lz]`` expands to an orthorhobic cuboid, and a 6-element
        *box* array represents a fully triclinic box in the same format as
        GSD and HOOMD: ``[Lx, Ly, Lz, xy, xz, yz]``.

    See Also:
        Tutorials:

        - :doc:`examples/01-Primitives/05-Box-geometry`
        - :doc:`examples/02-Advanced-topics/05-GSD-visualization`

    Note:
        The Box class is constructed from `spherocylinders <Cylinder>`, which
        can be modified individually. The convenience attributes ``box_radius``
        and ``box_color`` can be used to set the thickness and color of the
        entire box.
    """

    def __init__(self, scene, box, box_radius=0.5, box_color=[0, 0, 0]):

        super().__init__(scene=scene,
                         N=12,
                         material=material.Material(solid=1.0))
        self._box = self._from_box(box)
        self.points[:] = self._generate_points(self._box)

        self.box_radius = box_radius

        self.box_color = box_color

    def _from_box(self, box):
        """Duck type the box from a valid input.

        Boxes can be a number, list, dictionary, or object with attributes.
        """
        try:
            # Handles freud.box.Box and namedtuple
            Lx = box.Lx
            Ly = box.Ly
            Lz = getattr(box, 'Lz', 0)
            xy = getattr(box, 'xy', 0)
            xz = getattr(box, 'xz', 0)
            yz = getattr(box, 'yz', 0)
        except AttributeError:
            try:
                # Handle dictionary-like
                Lx = box['Lx']
                Ly = box['Ly']
                Lz = box.get('Lz', 0)
                xy = box.get('xy', 0)
                xz = box.get('xz', 0)
                yz = box.get('yz', 0)
            except (IndexError, KeyError, TypeError):
                try:
                    if not len(box) in [1, 3, 6]:
                        raise ValueError(
                            "List-like objects must have length 1, 3, or 6 to "
                            "be converted to a box.")
                    # Handle list-like
                    Lx = box[0]
                    Ly = box[0] if len(box) == 1 else box[1]
                    Lz = box[0] if len(box) == 1 else box[2]
                    xy, xz, yz = box[3:6] if len(box) == 6 else (0, 0, 0)
                except TypeError:
                    if isinstance(box, int) or isinstance(box, float):
                        # Handle int or float
                        Lx = box
                        Ly = box
                        Lz = box
                        xy = 0
                        xz = 0
                        yz = 0
                    else:
                        raise TypeError(f"unsupported box type {type(box)}")
        return (Lx, Ly, Lz, xy, xz, yz)

    def _generate_points(self, box):
        """Helper function to take a box and calculate the 12 edges."""
        Lx = box[0]
        Ly = box[1]
        Lz = box[2]
        xy = box[3]
        xz = box[4]
        yz = box[5]

        # Follow hoomd convention
        box_matrix = numpy.array([[Lx, xy * Ly, xz * Lz], [0, Ly, yz * Lz],
                                  [0, 0, Lz]])
        a_1, a_2, a_3 = box_matrix.T
        #           F--------------H
        #          /|             /|
        #         / |            / |
        #        D--+-----------E  |
        #        |  |           |  |
        #        |  |           |  |
        #        |  |           |  |
        #        |  C-----------+--G
        #        | /            | /
        #        |/             |/
        #        A--------------B
        # Translate A so that 0, 0, 0 is the center of the box
        A = -(a_1 + a_2 + a_3) / 2
        B = A + a_1
        C = A + a_2
        D = A + a_3
        E = A + a_1 + a_3
        F = A + a_2 + a_3
        G = A + a_1 + a_2
        H = A + a_1 + a_2 + a_3
        # Define all edges
        box_points = numpy.asarray([
            [A, B],
            [A, C],
            [A, D],
            [B, E],
            [B, G],
            [C, G],
            [C, F],
            [D, E],
            [D, F],
            [E, H],
            [F, H],
            [G, H],
        ])
        return box_points

    @property
    def box(self):
        """(1, ), (3, ), or (6, ) `numpy.ndarray` of ``float32``: Box\
            parameters.

        Set `box` to update the shape of the box.
        """
        return self._box

    @box.setter
    def box(self, box):
        self._box = self._from_box(box)
        self.points[:] = self._generate_points(self._box)

    @property
    def box_color(self):
        """(3, ) `numpy.ndarray` of ``float32``: Color of the box edges.

        Note:
            This property sets the color of the `material <Geometry.material>`.
        """
        return self.material.color

    @box_color.setter
    def box_color(self, color):
        self.material.color = color

    @property
    def box_radius(self):
        """(float): Radius of box edges."""
        return self.radius[:][0]

    @box_radius.setter
    def box_radius(self, radius):
        self.radius[:] = radius


class Polygon(Geometry):
    """Polygon geometry.

    Define a set of simple polygon primitives in the xy plane with individual
    positions, rotation angles, and colors.

    Args:
        scene (Scene): Add the geometry to this scene.

        vertices ((N_vert, 2) `numpy.ndarray` of ``float32``): Polygon vertices.

        position ((N, 2) `numpy.ndarray` of ``float32``): Position of each
            polygon.

        angle ((N, ) `numpy.ndarray` of ``float32``): Orientation angle of each
            polygon (in radians).

        color ((N, 3) `numpy.ndarray` of ``float32``): Color of each polygon.

        rounding_radius (float): Rounding radius for spheropolygons.

        N (int): Number of polygons in the geometry. If ``None``, determine
            *N* from *position*.

        material (material.Material): Define how light interacts with the
            geometry. When `None`, defaults to ``material.Material()``.

        outline_material (material.Material): Define how light interacts with \
            the geometry's outline. When `None`, defaults to solid black
            ``material.Material(solid=1, color=(0,0,0))``.

        outline_width (float): Width of the outline in scene units.

    See Also:
        Tutorials:

        - :doc:`examples/01-Primitives/04-Polygon-geometry`

    Hint:
        Avoid costly memory allocations and type conversions by specifying
        primitive properties in the appropriate array type.
    """

    def __init__(self,
                 scene,
                 vertices,
                 position=(0, 0),
                 angle=0,
                 color=(0, 0, 0),
                 rounding_radius=0,
                 N=None,
                 material=None,
                 outline_material=None,
                 outline_width=0.0):
        if N is None:
            N = len(position)

        self._geometry = scene.device.module.GeometryPolygon(
            scene._scene, vertices, rounding_radius, N)
        if material is None:
            self.material = globals()['material'].Material()
        else:
            self.material = material

        if outline_material is None:
            self.outline_material = globals()['material'].Material(solid=1,
                                                                   color=(0, 0,
                                                                          0))
        else:
            self.outline_material = outline_material

        self.outline_width = outline_width

        self.position[:] = position
        self.angle[:] = angle
        self.color[:] = color

        self.scene = scene
        self.scene.geometry.append(self)

    @property
    def position(self):
        """(N, 2) `Array`: The position of each polygon."""
        return util.Array(self._geometry.getPositionBuffer(), geom=self)

    @property
    def angle(self):
        """(N, ) `Array`: The rotation angle of each polygon (in radians)."""
        return util.Array(self._geometry.getAngleBuffer(), geom=self)

    @property
    def color(self):
        """(N, 2, 3) `Array`: The color of each polygon."""
        return util.Array(self._geometry.getColorBuffer(), geom=self)

    def get_extents(self):
        """Get the extents of the geometry.

        Returns:
            (3,2) `numpy.ndarray` of ``float32``: The lower left and\
                upper right corners of the scene.
        """
        pos = self.position[:]
        r = self._geometry.getRadius()
        res2d = numpy.array(
            [numpy.min(pos - r, axis=0),
             numpy.max(pos + r, axis=0)])
        res = numpy.array([[res2d[0][0], res2d[0][1], -1e-5],
                           [res2d[1][0], res2d[1][1], 1e-5]])

        return res


class Sphere(Geometry):
    """Sphere geometry.

    Define a set of sphere primitives with individual positions, radii, and
    colors.

    Args:
        scene (Scene): Add the geometry to this scene.

        position ((N, 3) `numpy.ndarray` of ``float32``):
            Position of each sphere.

        radius ((N, ) `numpy.ndarray` of ``float32``):
            Radius of each sphere.

        color ((N, 3) `numpy.ndarray` of ``float32``): Color of each sphere.

        N (int): Number of spheres in the geometry. If ``None``, determine *N*
            from *position*.

        material (material.Material): Define how light interacts with the
            geometry. When `None`, defaults to ``material.Material()``.

        outline_material (material.Material): Define how light interacts with \
            the geometry's outline. When `None`, defaults to solid black
            ``material.Material(solid=1, color=(0,0,0))``.

        outline_width (float): Width of the outline in scene units.

    See Also:
        Tutorials:

        - :doc:`examples/01-Primitives/00-Sphere-geometry`

    Hint:
        Avoid costly memory allocations and type conversions by specifying
        primitive properties in the appropriate array type.

    Tip:
        When all spheres are the same size, pass a single value for *radius* and
        numpy will broadcast it to all elements of the array.
    """

    def __init__(self,
                 scene,
                 position=(0, 0, 0),
                 radius=0.5,
                 color=(0, 0, 0),
                 N=None,
                 material=None,
                 outline_material=None,
                 outline_width=0.0):
        if N is None:
            N = len(position)

        self._geometry = scene.device.module.GeometrySphere(scene._scene, N)

        if material is None:
            self.material = globals()['material'].Material()
        else:
            self.material = material

        if outline_material is None:
            self.outline_material = globals()['material'].Material(solid=1,
                                                                   color=(0, 0,
                                                                          0))
        else:
            self.outline_material = outline_material

        self.outline_width = outline_width

        self.position[:] = position
        self.radius[:] = radius
        self.color[:] = color

        self.scene = scene
        self.scene.geometry.append(self)

    def get_extents(self):
        """Get the extents of the geometry.

        Returns:
            (3,2) `numpy.ndarray` of ``float32``: The lower left and\
                upper right corners of the scene.
        """
        pos = self.position[:]
        r = self.radius[:]
        r = r.reshape(len(r), 1)
        res = numpy.array(
            [numpy.min(pos - r, axis=0),
             numpy.max(pos + r, axis=0)])
        return res

    @property
    def position(self):
        """(N, 3) `Array`: The position of each sphere."""
        return util.Array(self._geometry.getPositionBuffer(), geom=self)

    @property
    def radius(self):
        """(N, ) `Array`: The radius of each sphere."""
        return util.Array(self._geometry.getRadiusBuffer(), geom=self)

    @property
    def color(self):
        """(N, 3) `Array`: The color of each sphere."""
        return util.Array(self._geometry.getColorBuffer(), geom=self)


class Mesh(Geometry):
    """Mesh geometry.

    Define a set of triangle mesh primitives with individual positions,
    orientations, and colors.

    Args:
        scene (Scene): Add the geometry to this scene.

        vertices ((3T, 3) `numpy.ndarray` of ``float32``):
            Vertices of the triangles, listed contiguously. Vertices 0,1,2
            define the first triangle, 3,4,5 define the second, and so on.

        color ((3T, 3) `numpy.ndarray` of ``float32``):
            Color of each vertex.

        position ((N, 3) `numpy.ndarray` of ``float32``):
            Position of each mesh instance.

        orientation ((N, 4) `numpy.ndarray` of ``float32``):
            Orientation of each mesh instance (as a quaternion).

        N (int): Number of mesh instances in the geometry. If ``None``,
            determine *N* from *position*.

        material (material.Material): Define how light interacts with the
            geometry. When `None`, defaults to ``material.Material()``.

        outline_material (material.Material): Define how light interacts with \
            the geometry's outline. When `None`, defaults to solid black
            ``material.Material(solid=1, color=(0,0,0))``.

        outline_width (float): Width of the outline in scene units.

    See Also:
        Tutorials:

        - :doc:`examples/01-Primitives/03-Mesh-geometry`

    Hint:
        Avoid costly memory allocations and type conversions by specifying
        primitive properties in the appropriate array type.
    """

    def __init__(self,
                 scene,
                 vertices,
                 position=(0, 0, 0),
                 orientation=(1, 0, 0, 0),
                 color=(0, 0, 0),
                 N=None,
                 material=None,
                 outline_material=None,
                 outline_width=0.0):
        if N is None:
            N = len(position)

        self.vertices = numpy.asarray(vertices, dtype=numpy.float32)
        self._geometry = scene.device.module.GeometryMesh(
            scene._scene, self.vertices, N)
        if material is None:
            self.material = globals()['material'].Material()
        else:
            self.material = material

        if outline_material is None:
            self.outline_material = globals()['material'].Material(solid=1,
                                                                   color=(0, 0,
                                                                          0))
        else:
            self.outline_material = outline_material

        self.outline_width = outline_width

        self.position[:] = position
        self.orientation[:] = orientation
        self.color[:] = color

        self.scene = scene
        self.scene.geometry.append(self)

    @property
    def position(self):
        """(N, 3) `Array`: The position of each mesh."""
        return util.Array(self._geometry.getPositionBuffer(), geom=self)

    @property
    def orientation(self):
        """(N, 4) `Array`: The orientation of each mesh."""
        return util.Array(self._geometry.getOrientationBuffer(), geom=self)

    @property
    def color(self):
        """(N, 3) `Array`: The color of each sphere."""
        return util.Array(self._geometry.getColorBuffer(), geom=self)

    def get_extents(self):
        """Get the extents of the geometry.

        Returns:
            (3,2) `numpy.ndarray` of ``float32``: The lower left and\
                upper right corners of the scene.
        """
        a = self.vertices[:, 0]
        b = self.vertices[:, 1]
        c = self.vertices[:, 2]
        r = numpy.array([
            numpy.min([
                numpy.min(a, axis=0),
                numpy.min(b, axis=0),
                numpy.min(c, axis=0)
            ],
                      axis=0),
            numpy.max([
                numpy.max(a, axis=0),
                numpy.max(b, axis=0),
                numpy.max(c, axis=0)
            ],
                      axis=0)
        ])

        pos = self.position[:]
        res = numpy.array(
            [numpy.min(pos + r[0], axis=0),
             numpy.max(pos + r[1], axis=0)])
        return res


class ConvexPolyhedron(Geometry):
    """Convex polyhedron geometry.

    Define a set of convex polyhedron primitives with individual positions,
    orientations, and colors.

    A convex polyhedron is defined by *P* outward facing planes (origin and
    normal vector) and a radius that encompass the shape. Use
    `convex_polyhedron_from_vertices` to construct this from the convex hull of
    a set of vertices.

    Args:
        scene (Scene): Add the geometry to this scene.

        polyhedron_info (Dict): A dictionary containing the face normals
            (``face_normal``), origins (``face_origin``), face colors
            (``face_color``), and the radius (``radius``)).

        position ((N, 3) `numpy.ndarray` of ``float32``):
            Position of each polyhedron instance.

        orientation ((N, 4) `numpy.ndarray` of ``float32``):
            Orientation of each polyhedron instance (as a quaternion).

        color ((N, 3) `numpy.ndarray` of ``float32``):
            Color of each polyhedron.

        N (int): Number of spheres in the geometry. If ``None``, determine *N*
            from *position*.

        material (material.Material): Define how light interacts with the
            geometry. When `None`, defaults to ``material.Material()``.

        outline_material (material.Material): Define how light interacts with \
            the geometry's outline. When `None`, defaults to solid black
            ``material.Material(solid=1, color=(0,0,0))``.

        outline_width (float): Width of the outline in scene units.

    See Also:
        Tutorials:

        - :doc:`examples/01-Primitives/02-Convex-polyhedron-geometry`

    Hint:
        Avoid costly memory allocations and type conversions by specifying
        primitive properties in the appropriate array type.
    """

    def __init__(self,
                 scene,
                 polyhedron_info,
                 position=(0, 0, 0),
                 orientation=(1, 0, 0, 0),
                 color=(0, 0, 0),
                 N=None,
                 material=None,
                 outline_material=None,
                 outline_width=0.0):
        if N is None:
            N = len(position)

        origins = polyhedron_info['face_origin']
        normals = polyhedron_info['face_normal']
        face_colors = polyhedron_info['face_color']
        r = polyhedron_info['radius']
        self._geometry = scene.device.module.GeometryConvexPolyhedron(
            scene._scene, origins, normals, face_colors, N, r)

        if material is None:
            self.material = globals()['material'].Material()
        else:
            self.material = material

        if outline_material is None:
            self.outline_material = globals()['material'].Material(solid=1,
                                                                   color=(0, 0,
                                                                          0))
        else:
            self.outline_material = outline_material

        self.outline_width = outline_width
        self._radius = r

        self.position[:] = position
        self.orientation[:] = orientation
        self.color[:] = color

        self.scene = scene
        self.scene.geometry.append(self)

    def get_extents(self):
        """Get the extents of the geometry.

        Returns:
            (3,2) `numpy.ndarray` of ``float32``: The lower left and\
                upper right corners of the scene.
        """
        pos = self.position[:]
        r = self._radius
        res = numpy.array(
            [numpy.min(pos - r, axis=0),
             numpy.max(pos + r, axis=0)])
        return res

    @property
    def position(self):
        """(N, 3) `Array`: The position of each polyhedron."""
        return util.Array(self._geometry.getPositionBuffer(), geom=self)

    @property
    def orientation(self):
        """(N, 4) `Array`: The orientation of each polyhedron."""
        return util.Array(self._geometry.getOrientationBuffer(), geom=self)

    @property
    def color(self):
        """(N, 3) `Array`: The color of each polyhedron."""
        return util.Array(self._geometry.getColorBuffer(), geom=self)

    @property
    def color_by_face(self):
        """float: Mix face colors with the per-polyhedron color.

        Set to 0 to color particles by the per-particle `color`. Set to 1 to
        color faces by the per-face color. Set to a value between 0 and 1 to
        blend between the two colors.
        """
        return self._geometry.getColorByFace()

    @color_by_face.setter
    def color_by_face(self, v):
        self._geometry.setColorByFace(v)



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test outline materials."""

import fresnel
from collections import namedtuple
import PIL
import conftest
import os
import pathlib

dir_path = pathlib.Path(os.path.realpath(__file__)).parent


def test_set_material(scene_hex_sphere_, generate=False):
    """Test that outline materials render properly."""
    geometry = scene_hex_sphere_.geometry[0]
    geometry.outline_width = 0.3
    geometry.outline_material = fresnel.material.Material(
        solid=0.0,
        color=fresnel.color.linear([1, 0, 0]),
        primitive_color_mix=0.0)
    assert geometry.outline_material.solid == 0.0
    assert geometry.outline_material.color == tuple(
        fresnel.color.linear([1, 0, 0]))
    assert geometry.outline_material.primitive_color_mix == 0.0

    buf_proxy = fresnel.preview(scene_hex_sphere_,
                                w=100,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_outline_material.test_set_material.png', 'wb'),
            'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:], dir_path / 'reference'
            / 'test_outline_material.test_set_material.png')


def test_solid(scene_hex_sphere_, generate=False):
    """Test the solid property."""
    geometry = scene_hex_sphere_.geometry[0]
    geometry.outline_width = 0.3
    geometry.outline_material.solid = 1.0
    assert geometry.outline_material.solid == 1.0

    buf_proxy = fresnel.preview(scene_hex_sphere_,
                                w=100,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_outline_material.test_solid.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_outline_material.test_solid.png')


def test_color(scene_hex_sphere_, generate=False):
    """Test the color property."""
    geometry = scene_hex_sphere_.geometry[0]
    geometry.outline_width = 0.3
    geometry.outline_material.color = fresnel.color.linear([0, 0, 1])
    assert geometry.outline_material.color == tuple(
        fresnel.color.linear([0, 0, 1]))

    buf_proxy = fresnel.preview(scene_hex_sphere_,
                                w=100,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_outline_material.test_color.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_outline_material.test_color.png')


def test_primitive_color_mix(scene_hex_sphere_, generate=False):
    """Test the color mix property."""
    geometry = scene_hex_sphere_.geometry[0]
    geometry.outline_width = 0.3
    geometry.outline_material = fresnel.material.Material(
        solid=1.0,
        color=fresnel.color.linear([1, 0, 0]),
        primitive_color_mix=1.0)

    geometry.color[0] = fresnel.color.linear([1, 0, 0])
    geometry.color[1] = fresnel.color.linear([0, 1, 0])
    geometry.color[2] = fresnel.color.linear([0, 0, 1])
    geometry.color[3] = fresnel.color.linear([1, 0, 1])
    geometry.color[4] = fresnel.color.linear([0, 1, 1])
    geometry.color[5] = fresnel.color.linear([0, 0, 0])

    buf_proxy = fresnel.preview(scene_hex_sphere_,
                                w=100,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_outline_material.test_primitive_color_mix.png',
                 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:], dir_path / 'reference'
            / 'test_outline_material.test_primitive_color_mix.png')


if __name__ == '__main__':
    struct = namedtuple("struct", "param")
    device = conftest.device(struct(('cpu', None)))

    scene = conftest.scene_hex_sphere(device)
    test_set_material(scene, generate=True)

    scene = conftest.scene_hex_sphere(device)
    test_solid(scene, generate=True)

    scene = conftest.scene_hex_sphere(device)
    test_color(scene, generate=True)

    scene = conftest.scene_hex_sphere(device)
    test_primitive_color_mix(scene, generate=True)



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Pytest fixtures and methods used across all fresnel tests."""

import pytest
import fresnel
import math
import PIL
import numpy
import os
import pathlib

dir_path = pathlib.Path(os.path.realpath(__file__)).parent

devices = []
if 'cpu' in fresnel.Device.available_modes:
    devices.append(('cpu', None))
    devices.append(('cpu', 1))

if 'gpu' in fresnel.Device.available_modes:
    devices.append(('gpu', 1))


def test_lights():
    """Create a set of lights to use in most tests."""
    lights = []
    phi1 = 1 * 45 * math.pi / 180
    theta1 = (90 - 20) * math.pi / 180
    lights.append(
        fresnel.light.Light(direction=(math.sin(theta1) * math.sin(phi1),
                                       math.cos(theta1),
                                       math.sin(theta1) * math.cos(phi1)),
                            color=(0.75, 0.75, 0.75),
                            theta=math.pi / 8))
    phi1 = -1 * 45 * math.pi / 180
    theta1 = (90) * math.pi / 180
    lights.append(
        fresnel.light.Light(direction=(math.sin(theta1) * math.sin(phi1),
                                       math.cos(theta1),
                                       math.sin(theta1) * math.cos(phi1)),
                            color=(0.1, 0.1, 0.1),
                            theta=math.pi / 2))
    return lights


def device(request):
    """Create a Device object for use in the tests."""
    mode = request.param[0]
    limit = request.param[1]

    dev = fresnel.Device(mode=mode, n=limit)

    return dev


def device_name(device):
    """Generate a string name for the device fixture."""
    name = device[0]
    if device[1] is None:
        name += '(all)'
    else:
        name += '(' + str(device[1]) + ')'
    return name


@pytest.fixture(scope='session', params=devices, ids=device_name)
def device_(request):
    """Pytest fixture parameterized over several devices."""
    return device(request)


def scene_hex_sphere(device):
    """Create a scene with six spheres arranged in a hexagon."""
    scene = fresnel.Scene(device, lights=test_lights())

    position = []
    for i in range(6):
        position.append([
            2 * math.cos(i * 2 * math.pi / 6),
            2 * math.sin(i * 2 * math.pi / 6), 0
        ])

    geometry = fresnel.geometry.Sphere(scene, position=position, radius=1.0)
    geometry.material = fresnel.material.Material(solid=0.0,
                                                  color=fresnel.color.linear(
                                                      [1, 0.874, 0.169]))
    geometry.outline_width = 0.12

    scene.camera = fresnel.camera.Orthographic(position=(0, 0, 10),
                                               look_at=(0, 0, 0),
                                               up=(0, 1, 0),
                                               height=6)

    return scene


@pytest.fixture(scope='function')
def scene_hex_sphere_(device_):
    """Pytest fixture to create a scene with six spheres."""
    return scene_hex_sphere(device_)


def assert_image_approx_equal(a, ref_file, tolerance=1.0):
    """Validate a rendered image.

    Check the output buffer *a* from fresnel matches the saved png file
    in *ref_file*.
    """
    im = PIL.Image.open(ref_file)
    im_arr = numpy.frombuffer(im.tobytes(), dtype=numpy.uint8)
    im_arr = im_arr.reshape((im.size[1], im.size[0], 4))

    # intelligently compare images first, ensure that they share a large
    # fraction of non-background pixels (this assumes that all image compare
    # tests use a background alpha = 0)
    a_selection = a[:, :, 3] > 0
    ref_selection = im_arr[:, :, 3] > 0
    assert numpy.sum(a_selection) > 3 / 4 * numpy.sum(ref_selection)

    # Now, compute the sum of the image difference squared, but only over those
    # pixels that are non-background in both images. This prevents a full 255
    # difference from showing up in a pixel that is present in one image but not
    # the other due to a round off error
    selection = a_selection * ref_selection
    a_float = numpy.array(a, dtype=numpy.float32)
    im_float = numpy.array(im_arr, dtype=numpy.float32)
    diff = numpy.array((a_float - im_float))
    msd = numpy.mean(diff[selection]**2)

    assert msd < tolerance


def _validate_or_generate_image(buf, name, generate, tolerance):
    """Validate or generate an image.

    When generate is True, save the image in *buf* to ``output/{name}.png``.
    When generate is False, compare the image in *buf* against
    the file ``reference/{name}.png``.
    """
    if generate:
        PIL.Image.fromarray(buf[:],
                            mode='RGBA').save(open(f'output/{name}.png', 'wb'),
                                              'png')
    else:
        assert_image_approx_equal(buf[:],
                                  dir_path / 'reference' / f'{name}.png',
                                  tolerance)


def check_preview_render(scene,
                         name,
                         w=150,
                         h=100,
                         tolerance=2,
                         anti_alias=False,
                         generate=False):
    """Preview render the given scene and check the image matches the ref."""
    buf_proxy = fresnel.preview(scene, w=w, h=h, anti_alias=anti_alias)

    _validate_or_generate_image(buf_proxy, name, generate, tolerance)


def check_pathtrace_render(scene,
                           name,
                           w=150,
                           h=100,
                           tolerance=2,
                           samples=64 * 40,
                           generate=False):
    """Pathtrace render the given scene and check the image matches the ref."""
    buf_proxy = fresnel.pathtrace(scene, w=w, h=h, samples=samples)

    _validate_or_generate_image(buf_proxy, name, generate, tolerance)



# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Test the Mesh geometry."""

import fresnel
from collections import namedtuple
import PIL
import conftest
import pytest
import os
import pathlib

dir_path = pathlib.Path(os.path.realpath(__file__)).parent


def scene_one_triangle(device):
    """Create a test scene with one triangle."""
    scene = fresnel.Scene(device, lights=conftest.test_lights())

    geometry = fresnel.geometry.Mesh(scene,
                                     vertices=[[-1, -1, 0], [1, -1, 0],
                                               [-1, 1, 0]],
                                     N=1)
    geometry.color[:] = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]

    geometry.material = fresnel.material.Material(color=fresnel.color.linear(
        [1.0, 0, 0]),
                                                  roughness=0.8,
                                                  specular=0.5,
                                                  primitive_color_mix=0.0,
                                                  solid=1)

    geometry.outline_material = fresnel.material.Material(
        color=fresnel.color.linear([0, 1.0, 0]),
        roughness=0.8,
        specular=0.5,
        primitive_color_mix=0.0,
        solid=1)

    geometry.orientation[:] = [1, 0, 0, 0]

    scene.camera = fresnel.camera.Orthographic(position=(0, 0, 20),
                                               look_at=(0, 0, 0),
                                               up=(0, 1, 0),
                                               height=2.1)

    return scene


@pytest.fixture(scope='function')
def scene_one_triangle_(device_):
    """Pytest fixture to create a test scene."""
    return scene_one_triangle(device_)


def scene_tetrahedra(device):
    """Create a test scene with a tetrahedron."""
    scene = fresnel.Scene(device, lights=conftest.test_lights())

    verts = [(1, 1, 1), (1, -1, -1), (-1, 1, -1), (-1, -1, 1)]
    triangles = [
        verts[0], verts[1], verts[2], verts[2], verts[1], verts[3], verts[2],
        verts[3], verts[0], verts[1], verts[0], verts[3]
    ]

    geometry = fresnel.geometry.Mesh(scene, vertices=triangles, N=4)
    geometry.color[0:2, :] = [0.9, 0, 0]
    geometry.color[3:5, :] = [0, 0.9, 0]
    geometry.color[6:8, :] = [0, 0, 0.9]
    geometry.color[9:11, :] = [0.9, 0, 0.9]

    geometry.material = fresnel.material.Material(color=fresnel.color.linear(
        [1.0, 0, 0]),
                                                  roughness=0.8,
                                                  specular=0.5,
                                                  primitive_color_mix=1.0,
                                                  solid=0)

    geometry.position[:] = [[-2, -2, 0], [2, -2, 0], [2, 2, 0], [-2, 2, 0]]

    geometry.orientation[:] = [
        [0.03723867, 0.38927173, -0.73216521, -0.55768711],
        [-0.32661186, 0.43644863, -0.09899935, 0.83248808],
        [0.25624845, 0.32632096, -0.11995704, -0.9019211],
        [-0.78025512, -0.12102377, 0.24947819, 0.56063877]
    ]

    scene.camera = fresnel.camera.Orthographic(position=(0, 0, -20),
                                               look_at=(0, 0, 0),
                                               up=(0, 1, 0),
                                               height=7.5)

    return scene


@pytest.fixture(scope='function')
def scene_tetrahedra_(device_):
    """Pytest fixture to create a test scene."""
    return scene_tetrahedra(device_)


def test_render(scene_one_triangle_, generate=False):
    """Test that Mesh renders properly."""
    buf_proxy = fresnel.preview(scene_one_triangle_,
                                w=100,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_mesh.test_render.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_mesh.test_render.png')


def test_outline(scene_one_triangle_, generate=False):
    """Test that outlines render properly."""
    geometry = scene_one_triangle_.geometry[0]
    geometry.outline_width = 0.1

    buf_proxy = fresnel.preview(scene_one_triangle_,
                                w=100,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_mesh.test_outline.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_mesh.test_outline.png')


def test_color_interp(scene_one_triangle_, generate=False):
    """Test that colors are interpolated between vertices."""
    geometry = scene_one_triangle_.geometry[0]
    geometry.material.primitive_color_mix = 1.0

    buf_proxy = fresnel.preview(scene_one_triangle_,
                                w=100,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_mesh.test_color_interp.png', 'wb'),
            'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_mesh.test_color_interp.png')


def test_multiple(scene_tetrahedra_, generate=False):
    """Test that Mesh supports multiple instances."""
    buf_proxy = fresnel.preview(scene_tetrahedra_,
                                w=100,
                                h=100,
                                anti_alias=False)

    if generate:
        PIL.Image.fromarray(buf_proxy[:], mode='RGBA').save(
            open('output/test_geometry_mesh.test_multiple.png', 'wb'), 'png')
    else:
        conftest.assert_image_approx_equal(
            buf_proxy[:],
            dir_path / 'reference' / 'test_geometry_mesh.test_multiple.png')


if __name__ == '__main__':
    struct = namedtuple("struct", "param")
    device = conftest.device(struct(('cpu', None)))

    scene = scene_one_triangle(device)
    test_render(scene, generate=True)

    scene = scene_one_triangle(device)
    test_outline(scene, generate=True)

    scene = scene_one_triangle(device)
    test_color_interp(scene, generate=True)

    scene = scene_tetrahedra(device)
    test_multiple(scene, generate=True)
